/* ARCADEmini USB driver for Windows
 * version 1.1.0 date 2005-02-10
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <wdm.h>
#include <stdarg.h>
#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "AmUsbInternal.h"
#include "AmUsb.h"



//*** support functions

//free a communication block
static void AmUsbCommBlockFree( struct tAmUsbCommBlock * pCommBlock )
{
	int Step;

	//communication block is still linked
	if( pCommBlock->Linked )
	{
		//remove pointer to wait event
		pCommBlock->pWaitEvent = NULL;
		//set free flag to free the block automatically after completion
		pCommBlock->FreeFlag = 1;
		//do nothing else
		return;
	}

	//free buffers of communication steps
	for( Step = 0; Step < pCommBlock->StepCnt; Step++ )
	{
		if( pCommBlock->Steps[Step].pOutBuffer != NULL )
			ExFreePool( pCommBlock->Steps[Step].pOutBuffer );
		if( pCommBlock->Steps[Step].pFilterBuffer != NULL )
			ExFreePool( pCommBlock->Steps[Step].pFilterBuffer );
		if( pCommBlock->Steps[Step].pMaskBuffer != NULL )
			ExFreePool( pCommBlock->Steps[Step].pMaskBuffer );
		if( pCommBlock->Steps[Step].pInBuffer != NULL )
			ExFreePool( pCommBlock->Steps[Step].pInBuffer );
	}

	//free array with communication steps
	ExFreePool( pCommBlock->Steps );

	//free communication block
	ExFreePool( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbCommBlockFree: freed communication block %p\r\n", pCommBlock ) );
}



//*** bulk output interrupt functions

//callback-procedure for USB bulk output
//pContext is pointer to private device data
//(interrupt context)
//ATTENTION: pDeviceObject is NULL - do not use
static NTSTATUS AmUsbOutCallback( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp, IN PVOID pContext )
{
	struct tAmUsbDev * pDev;
	int Len, I, J;
	char DbgText[256];

	AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbOutCallback\r\n" ) );

	//get private device data
	pDev = (struct tAmUsbDev *)pContext;

	//IRP-CANCEL-PROCEDURE (described in AmUsbInternal.h)
	if( InterlockedExchangePointer( &pDev->pOutIrp, NULL ) == NULL )
	{
		if( InterlockedExchange( &pDev->OutCancelled, 1 ) == 0 )
		{
			//bulk output is no more active
			pDev->OutActive = 0;
			AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbOutCallback - nothing to do\r\n" ) );
			//the Win98DDK-Help-File says we must return this when using driver allocated IRP
			return STATUS_MORE_PROCESSING_REQUIRED;
		}
	}

	//get length of data
	Len = pDev->pOutUrb->UrbBulkOrInterruptTransfer.TransferBufferLength;

	//error
	if( !NT_SUCCESS( pIrp->IoStatus.Status ) )
	{
		//print error message
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbOutCallback: bulk output failed with 0x%08X\r\n", pIrp->IoStatus.Status ) );
	}
	//no error
	else
	{
		//print bulk message with data
		DbgText[0] = 0;
		for( I = 0, J = 0; I < Len && J < sizeof( DbgText ) - 4; I++, J += 3 )
			sprintf( DbgText + J, "%02X ", pDev->pOutBuffer[I] );
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbOutCallback: bulk output done (L=%d): %s\r\n", Len, DbgText ) );
	}

	//bulk output is no more active
	if( InterlockedExchange( &pDev->OutActive, 0 ) ) //it was active
	{
		//free IRP
		IoFreeIrp( pIrp );
		//free URB
		ExFreePool( pDev->pOutUrb );
		pDev->pOutUrb = NULL;
		//free buffer
		ExFreePool( pDev->pOutBuffer );
		pDev->pOutBuffer = NULL;
	}

	//set flag for execution of AmUsbCommBlockSend
	pDev->TaskCommBlockSend = 1;
	//schedule task for things that cannot be done in interrupt
	if( ! pDev->Unload )
		KeInsertQueueDpc( &pDev->DpcObject, NULL, NULL ); //no system arguments

	AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbOutCallback\r\n" ) );

	//the Win98DDK-Help-File says we must return this when using driver allocated IRP
	return STATUS_MORE_PROCESSING_REQUIRED;
}



//*** bulk input interrupt functions

//process a bulk in message
//(may be called in interrupt context with device locked by someone else)
static void AmUsbInMsg( struct tAmUsbDev * pDev, unsigned char * pMsg, int MsgLen )
{
	struct tAmUsbCommBlock * pCommBlock;
	int Step, I;

	//no first communication block
	if( pDev->CommBlockList.pNext == &pDev->CommBlockList )
		return; //do nothing

	//get pointer to first communication block
	pCommBlock = pDev->CommBlockList.pNext;

	//at least one input communication step left to do
	if( pCommBlock->InStep < pCommBlock->StepCnt )
	{
		//filter and mask buffer present
		Step = pCommBlock->InStep;
		if( pCommBlock->Steps[Step].pFilterBuffer != NULL
		 && pCommBlock->Steps[Step].pMaskBuffer != NULL )
		{
			//match message against next step
			if( pCommBlock->Steps[Step].InLen != MsgLen ) //wrong length
				return; //do nothing
			for( I = 0; I < MsgLen; I++ )
				if( pCommBlock->Steps[Step].pFilterBuffer[I] !=
						(pMsg[I] & pCommBlock->Steps[Step].pMaskBuffer[I]) )
					break; //byte does not match
			if( I < MsgLen ) //some byte does not match
				return; //do nothing
		}

		//copy recevied message into input buffer
		if( pCommBlock->Steps[Step].pInBuffer != NULL )
			memcpy( pCommBlock->Steps[Step].pInBuffer, pMsg, MsgLen );

		//increment number of current input step
		pCommBlock->InStep++;
	}

	//communication block done
	if( (pCommBlock->OutStep >= pCommBlock->StepCnt
		&& pCommBlock->InStep >= pCommBlock->StepCnt) )
	{
		//delete timeout timer
		if( pDev->TimeoutTimerActive )
		{
			KeCancelTimer( &pDev->TimeoutTimer );
			KeRemoveQueueDpc( &pDev->TimeoutDpc );
			pDev->TimeoutTimerActive = 0;
		}

		//set flag for execution of AmUsbCommBlockRecv
		pDev->TaskCommBlockRecv = 1;
		//schedule task for things that cannot be done in interrupt
		if( ! pDev->Unload )
			KeInsertQueueDpc( &pDev->DpcObject, NULL, NULL ); //no system arguments
	}
}

//callback-procedure for USB bulk input
//pContext is pointer to private device data
//(interrupt context)
//ATTENTION: pDeviceObject is NULL - do not use
static NTSTATUS AmUsbInCallback( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp, IN PVOID pContext )
{
	struct tAmUsbDev * pDev;
	int Len, I, J;
	char DbgText[256];

	AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbInCallback\r\n" ) );

	//get private device data
	pDev = (struct tAmUsbDev *)pContext;

	//IRP-CANCEL-PROCEDURE (described in AmUsbInternal.h)
	if( InterlockedExchangePointer( &pDev->pInIrp, NULL ) == NULL )
	{
		if( InterlockedExchange( &pDev->InCancelled, 1 ) == 0 )
		{
			//bulk input is no more active
			pDev->InActive = 0;
			AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbInCallback - nothing to do\r\n" ) );
			//the Win98DDK-Help-File says we must return this when using driver allocated IRP
			return STATUS_MORE_PROCESSING_REQUIRED;
		}
	}

	//get length of data
	Len = pDev->pInUrb->UrbBulkOrInterruptTransfer.TransferBufferLength;

	//error
	if( !NT_SUCCESS( pIrp->IoStatus.Status ) )
	{
		//print error message
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbInCallback: bulk input failed with 0x%08X\r\n", pIrp->IoStatus.Status ) );
	}
	//no error
	else
	{
		//print bulk message with data
		DbgText[0] = 0;
		for( I = 0, J = 0; I < Len && J < sizeof( DbgText ) - 4; I++, J += 3 )
			sprintf( DbgText + J, "%02X ", pDev->pInBuffer[I] );
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbInCallback: bulk input done (L=%d): %s\r\n", Len, DbgText ) );

		//process the message
		//(device may be locked by someone else, we are in interrupt context)
		AmUsbInMsg( pDev, pDev->pInBuffer, Len );

		//schedule start of next bulk input
		//set flag for execution of AmUsbInStart
		pDev->TaskInStart = 1;
		//schedule task for things that cannot be done in interrupt
		if( ! pDev->Unload )
			KeInsertQueueDpc( &pDev->DpcObject, NULL, NULL ); //no system arguments
	}

	//bulk input is no more active
	if( InterlockedExchange( &pDev->InActive, 0 ) ) //it was active
	{
		//free IRP
		IoFreeIrp( pIrp );
		//free URB
		ExFreePool( pDev->pInUrb );
		pDev->pInUrb = NULL;
		//free buffer
		ExFreePool( pDev->pInBuffer );
		pDev->pInBuffer = NULL;
	}

	AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbInCallback\r\n" ) );

	//the Win98DDK-Help-File says we must return this when using driver allocated IRP
	return STATUS_MORE_PROCESSING_REQUIRED;
}

//callback-procedure for USB bulk timeout
//(interrupt context)
static void AmUsbTimeout( IN PKDPC pDpc, IN PVOID pContext, IN PVOID pSysArg1, IN PVOID SysArg2 )
{
	struct tAmUsbDev * pDev;
	struct tAmUsbCommBlock * pCommBlock;

	AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbTimeout\r\n" ) );

	//get pointer to private device data
	pDev = (struct tAmUsbDev *)pContext;

	//timeout timer no more active
	pDev->TimeoutTimerActive = 0;

	//no first communication block
	if( pDev->CommBlockList.pNext == &pDev->CommBlockList )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbTimeout - nothing to do\r\n" ) );
		return; //do nothing
	}

	//get pointer to first communication block
	pCommBlock = pDev->CommBlockList.pNext;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbTimeout: comunication block %p timed out\r\n", pCommBlock ) );

	//set timeout flag
	pCommBlock->Timeout = 1;

	//set flag for execution of AmUsbCommBlockRecv
	pDev->TaskCommBlockRecv = 1;
	//schedule task for things that cannot be done in interrupt
	if( ! pDev->Unload )
		KeInsertQueueDpc( &pDev->DpcObject, NULL, NULL ); //no system arguments

	AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbTimeout\r\n" ) );
}



//*** support functions for non-interrupt time (used in tasks)

//cancel USB bulk output
static void AmUsbOutCancel( struct tAmUsbDev * pDev )
{
	PIRP pOutIrp;

	AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbOutCancel\r\n" ) );

	//bulk output is not active
	if( !pDev->OutActive )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbOutCancel - not active\r\n" ) );
		//do nothing
		return;
	}

	//IRP-CANCEL-PROCEDURE (described in AmUsbInternal.h)
	pOutIrp = (PIRP)InterlockedExchangePointer( &pDev->pOutIrp, NULL );
	if( pOutIrp == NULL )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbOutCancel - no OutIrp\r\n" ) );
		return;
	}
	IoCancelIrp( pOutIrp );
	if( InterlockedExchange( &pDev->OutCancelled, 1 ) == 0 )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbOutCancel - not cancelled\r\n" ) );
		return;
	}

	//print info message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbOutCancel: bulk output cancelled\r\n" ) );

	//free IRP
	IoFreeIrp( pOutIrp );
	//free URB
	ExFreePool( pDev->pOutUrb );
	pDev->pOutUrb = NULL;
	//free buffer
	ExFreePool( pDev->pOutBuffer );
	pDev->pOutBuffer = NULL;

	AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbOutCancel\r\n" ) );
}

//cancel USB bulk input
static void AmUsbInCancel( struct tAmUsbDev * pDev )
{
	PIRP pInIrp;

	AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbInCancel\r\n" ) );

	//bulk input is not active
	if( !pDev->InActive )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbInCancel - not active\r\n" ) );
		//do nothing
		return;
	}

	//IRP-CANCEL-PROCEDURE (described in AmUsbInternal.h)
	pInIrp = (PIRP)InterlockedExchangePointer( &pDev->pInIrp, NULL );
	if( pInIrp == NULL )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbInCancel - no InIrp\r\n" ) );
		return;
	}
	IoCancelIrp( pInIrp );
	if( InterlockedExchange( &pDev->InCancelled, 1 ) == 0 )
	{
		AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbInCancel - not cancelled\r\n" ) );
		return;
	}

	//print info message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbInCancel: bulk input cancelled\r\n" ) );

	//free IRP
	IoFreeIrp( pInIrp );
	//free URB
	ExFreePool( pDev->pInUrb );
	pDev->pInUrb = NULL;
	//free buffer
	ExFreePool( pDev->pInBuffer );
	pDev->pInBuffer = NULL;

	AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbInCancel\r\n" ) );
}

//start bulk output
//pData must point to a buffer allocated with ExAllocatePool( NonPagedPool, DataLen )
//this buffer must never be freed by the caller
//returns -1 on error, 0 on success, 1 if bulk output already active, 2 is device is being stopped
//(must be called with device locked)
static int AmUsbOutStart( struct tAmUsbDev * pDev, unsigned char * pData, int DataLen )
{
	CCHAR StackSize;
	NTSTATUS Status;
	PIO_STACK_LOCATION pNextStack;
	int I, J;
	char DbgText[256];

	//no pointer to device extension
	if( pDev->pDeviceExtension == NULL || pDev->Unload )
	{
		//free buffer allocated by our caller
		ExFreePool( pData );
		//this device is being stopped - do nothing
		return 2;
	}

	//check buffer allocated by caller
	AmUsbAssert( pData != NULL );
	AmUsbAssert( DataLen >= 0 );

	//bulk output is already active
	if( pDev->OutActive )
	{
		//print info message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbOutStart: bulk output already active\r\n" ) );
		//free buffer allocated by our caller
		ExFreePool( pData );
		//do not start bulk output twice
		return 1;
	}

	//no output IRP, URB or buffer may be in use here
	AmUsbAssert( pDev->pOutIrp == NULL );
	AmUsbAssert( pDev->pOutUrb == NULL );
	AmUsbAssert( pDev->pOutBuffer == NULL );

	//allocate an IRP
	StackSize = pDev->pDeviceExtension->pTopOfStackDeviceObject->StackSize;
	pDev->pOutIrp = IoAllocateIrp( StackSize, FALSE );
	if( pDev->pOutIrp == NULL )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbOutStart: could not allocate IRP\r\n" ) );
		//free buffer allocated by our caller
		ExFreePool( pData );
		//error
		return -1;
	}

	//allocate an URB
	pDev->pOutUrb = (PURB)ExAllocatePool( NonPagedPool, sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER ) );
	if( pDev->pOutUrb == NULL )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbOutStart: could not allocate URB\r\n" ) );
		//free buffer allocated by our caller
		ExFreePool( pData );
		//free IRP
		IoFreeIrp( pDev->pOutIrp );
		pDev->pOutIrp = NULL;
		//error
		return -1;
	}

	//buffer allocated by caller gets our output buffer
	pDev->OutBufferSize = DataLen;
	if( pDev->OutBufferSize > pDev->OutMaxPacketSize )
		pDev->OutBufferSize = pDev->OutMaxPacketSize;
	pDev->pOutBuffer = pData;

	//initialize the IRP
	IoInitializeIrp( pDev->pOutIrp, IoSizeOfIrp( StackSize ), StackSize );

	//set completion routine
	IoSetCompletionRoutine( pDev->pOutIrp, AmUsbOutCallback,
			pDev, //context is pointer to private device data
			TRUE, TRUE, TRUE ); //call on success, on error and on cancel

	//set up the IRP for an internal control request
	pNextStack = IoGetNextIrpStackLocation( pDev->pOutIrp );
	pNextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
	pNextStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_INTERNAL_USB_SUBMIT_URB;
	pNextStack->Parameters.Others.Argument1 = pDev->pOutUrb;
	//the IRP is not yet cancelled
	pDev->pOutIrp->Cancel = FALSE;

	//initialize for IRP-CANCEL-PROCEDURE (described in AmUsbInternal.h)
	pDev->OutCancelled = 0;

	//set up URB for bulk output
	UsbBuildInterruptOrBulkTransferRequest( pDev->pOutUrb, sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER ), //sizeof( URB ) does not work here - this should be safe, because URB is larger than struct _URB_BULK...
			pDev->OutPipeHandle, //handle of the bulk output pipe
			pDev->pOutBuffer, //output buffer
			NULL, //we are not using MDL
			pDev->OutBufferSize, //length of transfer buffer
			0, //direction=output
			NULL ); //no linked URB

	//bulk output will be active very soon
	//(do this here, because callback procedure could be called while still not returned from IoCallDriver)
	pDev->OutActive = 1;

	//call the class driver to perform the operation
	Status = IoCallDriver( pDev->pDeviceExtension->pTopOfStackDeviceObject, pDev->pOutIrp );

	//some error occured
	if( !NT_SUCCESS( Status ) )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbOutStart: bulk output could not be started (Status=0x%08X)\r\n", Status ) );
		//bulk output is not active
		if( InterlockedExchange( &pDev->OutActive, 0 ) ) //it was still active
		{
			//free buffer
			ExFreePool( pDev->pOutBuffer );
			pDev->pOutBuffer = NULL;
			//free URB
			ExFreePool( pDev->pOutUrb );
			pDev->pOutUrb = NULL;
			//free IRP
			IoFreeIrp( pDev->pOutIrp );
			pDev->pOutIrp = NULL;
		}
		//return error
		return -1;
	}

	//print info that bulk output was started
	DbgText[0] = 0;
	for( I = 0, J = 0; I < DataLen && J < sizeof( DbgText ) - 4; I++, J += 3 )
		sprintf( DbgText + J, "%02X ", pData[I] );
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbOutStart: bulk output started (L=%d): %s\r\n", DataLen, DbgText ) );

	//return success
	return 0;
}

//start bulk input
//returns -1 on error, 0 on success, 1 if bulk input already active, 2 is device is being stopped
//(must be called with device locked)
static int AmUsbInStart( struct tAmUsbDev * pDev )
{
	CCHAR StackSize;
	NTSTATUS Status;
	PIO_STACK_LOCATION pNextStack;

	//no pointer to device extension
	if( pDev->pDeviceExtension == NULL || pDev->Unload )
		//this device is being stopped - do nothing
		return 2;

	//bulk input is already active
	if( pDev->InActive )
	{
		//print info message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbInStart: bulk input already active\r\n" ) );
		//do not start bulk input twice
		return 1;
	}

	//no input IRP, URB or buffer may be in use here
	AmUsbAssert( pDev->pInIrp == NULL );
	AmUsbAssert( pDev->pInUrb == NULL );
	AmUsbAssert( pDev->pInBuffer == NULL );

	//allocate an IRP
	StackSize = pDev->pDeviceExtension->pTopOfStackDeviceObject->StackSize;
	pDev->pInIrp = IoAllocateIrp( StackSize, FALSE );
	if( pDev->pInIrp == NULL )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbInStart: could not allocate IRP\r\n" ) );
		//error
		return -1;
	}

	//allocate an URB
	pDev->pInUrb = (PURB)ExAllocatePool( NonPagedPool, sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER ) );
	if( pDev->pInUrb == NULL )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbInStart: could not allocate URB\r\n" ) );
		//free IRP
		IoFreeIrp( pDev->pInIrp );
		pDev->pInIrp = NULL;
		//error
		return -1;
	}

	//allocate a buffer
	pDev->InBufferSize = pDev->InMaxPacketSize;
	pDev->pInBuffer = (unsigned char *)ExAllocatePool( NonPagedPool, pDev->InBufferSize );
	if( pDev->pInBuffer == NULL )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbInStart: could not allocate buffer\r\n" ) );
		//free URB
		ExFreePool( pDev->pInUrb );
		pDev->pInUrb = NULL;
		//free IRP
		IoFreeIrp( pDev->pInIrp );
		pDev->pInIrp = NULL;
		//error
		return -1;
	}

	//initialize the IRP
	IoInitializeIrp( pDev->pInIrp, IoSizeOfIrp( StackSize ), StackSize );
	
	//set completion routine
	IoSetCompletionRoutine( pDev->pInIrp, AmUsbInCallback,
			pDev, //context is pointer to private device data
			TRUE, TRUE, TRUE ); //call on success, on error and on cancel

	//set up the IRP for an internal control request
	pNextStack = IoGetNextIrpStackLocation( pDev->pInIrp );
	pNextStack->MajorFunction = IRP_MJ_INTERNAL_DEVICE_CONTROL;
	pNextStack->Parameters.DeviceIoControl.IoControlCode = IOCTL_INTERNAL_USB_SUBMIT_URB;
	pNextStack->Parameters.Others.Argument1 = pDev->pInUrb;
	//the IRP is not yet cancelled
	pDev->pInIrp->Cancel = FALSE;

	//initialize for IRP-CANCEL-PROCEDURE (described in AmUsbInternal.h)
	pDev->InCancelled = 0;

	//set up URB for bulk input
	UsbBuildInterruptOrBulkTransferRequest( pDev->pInUrb, sizeof( struct _URB_BULK_OR_INTERRUPT_TRANSFER ), //sizeof( URB ) does not work here - this should be safe, because URB is larger than struct _URB_BULK...
			pDev->InPipeHandle, //handle of the bulk input pipe
			pDev->pInBuffer, //input buffer
			NULL, //we are not using MDL
			pDev->InBufferSize, //length of transfer buffer
			USBD_TRANSFER_DIRECTION_IN | USBD_SHORT_TRANSFER_OK, //direction=input, shorter than maxlen packets are okay
			NULL ); //no linked URB

	//bulk input will be active very soon
	//(do this here, because callback procedure could be called while still not returned from IoCallDriver)
	pDev->InActive = 1;

	//call the class driver to perform the operation
	Status = IoCallDriver( pDev->pDeviceExtension->pTopOfStackDeviceObject, pDev->pInIrp );

	//some error occured
	if( !NT_SUCCESS( Status ) )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbInStart: bulk input could not be started (Status=0x%08X)\r\n", Status ) );
		//bulk input is not active
		if( InterlockedExchange( &pDev->InActive, 0 ) ) //it was still active
		{
			//free buffer
			ExFreePool( pDev->pInBuffer );
			pDev->pInBuffer = NULL;
			//free URB
			ExFreePool( pDev->pInUrb );
			pDev->pInUrb = NULL;
			//free IRP
			IoFreeIrp( pDev->pInIrp );
			pDev->pInIrp = NULL;
		}
		//return error
		return -1;
	}

	//print info that bulk input was started
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbInStart: bulk input started (maxLen=%d)\r\n", pDev->InMaxPacketSize ) );

	//return success
	return 0;
}



//*** functions to do some things that cannot be done in interrupt

//try to send next step of communication block
//(must be called with device locked)
static void AmUsbCommBlockSend( struct tAmUsbDev * pDev )
{
	struct tAmUsbCommBlock * pCommBlock;
	int Step, Len;
	unsigned char * pBuffer;
	LARGE_INTEGER Time;

	//bulk out is already active
	if( pDev->OutActive )
		//do nothing
		return;

	//no first communication block
	if( pDev->CommBlockList.pNext == &pDev->CommBlockList )
		return; //do nothing

	//get pointer to first communication block
	pCommBlock = pDev->CommBlockList.pNext;

	//all output communication steps done
	if( pCommBlock->OutStep >= pCommBlock->StepCnt )
		return; //do nothing

	//allocate a buffer of maximum length
	pBuffer = (unsigned char *)ExAllocatePool( NonPagedPool, pDev->OutMaxPacketSize );
	if( pBuffer == NULL ) //no memory available
	{
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbCommBlockSend: out of memory\r\n" ) );
		return; //do nothing
	}

	//copy next step into bulk output buffer
	Step = pCommBlock->OutStep;
	Len = min( pCommBlock->Steps[Step].OutLen, pDev->OutMaxPacketSize );
	memcpy( pBuffer, pCommBlock->Steps[Step].pOutBuffer, Len );

	//delete old timeout timer
	if( pDev->TimeoutTimerActive )
	{
		KeCancelTimer( &pDev->TimeoutTimer );
		KeRemoveQueueDpc( &pDev->TimeoutDpc );
		pDev->TimeoutTimerActive = 0;
	}
	//create timeout timer
	Time = RtlConvertLongToLargeInteger( AmUsbTimeoutTime );
	KeSetTimer( &pDev->TimeoutTimer, Time, &pDev->TimeoutDpc );
	pDev->TimeoutTimerActive = 1;

	//increment number of current output step
	pCommBlock->OutStep++;

	//start bulk output
	//buffer must never be freed by us, because AmUsbOutStart takes care of it
	AmUsbOutStart( pDev, pBuffer, Len );
}

//post-process received steps of communication block
//(must be called with device locked)
static void AmUsbCommBlockRecv( struct tAmUsbDev * pDev )
{
	struct tAmUsbCommBlock * pCommBlock;

	//no first communication block
	if( pDev->CommBlockList.pNext == &pDev->CommBlockList )
		return; //do nothing

	//get pointer to first communication block
	pCommBlock = pDev->CommBlockList.pNext;

	//communication block done
	if( (pCommBlock->OutStep >= pCommBlock->StepCnt
	  && pCommBlock->InStep >= pCommBlock->StepCnt)
	 || pCommBlock->Timeout )
	{
		//remove first communication block from list
		pDev->CommBlockList.pNext = pDev->CommBlockList.pNext->pNext;
		pDev->CommBlockList.pNext->pPrev = &pDev->CommBlockList;
		pCommBlock->Linked = 0;

		//debug message
		AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbCommBlockRecv: completed communication block %p\r\n", pCommBlock ) );

		//signal wait event of this communication block
		//(we are at DISPATCH_LEVEL, pointer cannot be stolen between if and KeSetEvent)
		if( pCommBlock->pWaitEvent != NULL )
			KeSetEvent( pCommBlock->pWaitEvent, IO_NO_INCREMENT, FALSE );

		//free communication block
		if( pCommBlock->FreeFlag )
			AmUsbCommBlockFree( pCommBlock );

		//set flag for execution of AmUsbCommBlockSend
		pDev->TaskCommBlockSend = 1;
		//schedule task for things that cannot be done in interrupt
		if( ! pDev->Unload )
			KeInsertQueueDpc( &pDev->DpcObject, NULL, NULL ); //no system arguments
	}
}

//task to do some things that cannot be done in interrupt
//pContext is pointer to private device data
static void AmUsbTask( IN PKDPC pDpc, IN PVOID pContext, IN PVOID pSysArg1, IN PVOID SysArg2 )
{
	struct tAmUsbDev * pDev;
	KIRQL OldIrql;

	//get pointer to private device data
	pDev = (struct tAmUsbDev *)pContext;

	//before waiting for lock, peek at flags to see if we have got something to do
	if( ! pDev->TaskInStart && ! pDev->TaskCommBlockSend && ! pDev->TaskCommBlockRecv )
		//nothing to do
		return;

	//debug message
	AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbTask\r\n" ) );

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );

	//call functions whose flag is set and clear the flag
	if( InterlockedExchange( &pDev->TaskInStart, 0 ) )
		AmUsbInStart( pDev );
	if( InterlockedExchange( &pDev->TaskCommBlockSend, 0 ) )
		AmUsbCommBlockSend( pDev );
	if( InterlockedExchange( &pDev->TaskCommBlockRecv, 0 ) )
		AmUsbCommBlockRecv( pDev );

	//unlock private device data
	//(this restores the original IRQL afterwards)
	KeReleaseSpinLock( &pDev->Lock, OldIrql );

	//debug message
	AmUsbDebug( DBGLVL_HIGH, ( "exiting from AmUsbTask\r\n" ) );
}



//*** support functions for non-interrupt-time

//allocate a communication block
static struct tAmUsbCommBlock * AmUsbCommBlockNew( int StepCnt, char * pOutLens, char * pInLens )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I;
	
	//invalid number of steps
	if( StepCnt <= 0 )
		return NULL;

	//allocate communication block
	pCommBlock = (struct tAmUsbCommBlock *)ExAllocatePool( NonPagedPool, sizeof( struct tAmUsbCommBlock ) );
	if( pCommBlock == NULL )
		return NULL;
	//initialize communication block
	pCommBlock->Linked = 0;
	pCommBlock->StepCnt = StepCnt;
	pCommBlock->OutStep = 0;
	pCommBlock->InStep = 0;
	pCommBlock->Timeout = 0;
	pCommBlock->pWaitEvent = NULL;
	pCommBlock->FreeFlag = 0;

	//allocate communication steps
	pCommBlock->Steps = (struct tAmUsbCommStep *)ExAllocatePool( NonPagedPool, StepCnt * sizeof( struct tAmUsbCommStep ) );
	if( pCommBlock->Steps == NULL )
	{
		ExFreePool( pCommBlock );
		return NULL;
	}
	//initialize steps
	for( I = 0; I < StepCnt; I++ )
	{
		pCommBlock->Steps[I].OutLen = max( 1, (int)(pOutLens[I]) );
		pCommBlock->Steps[I].pOutBuffer = NULL;
		pCommBlock->Steps[I].InLen = max( 0, (int)(pInLens[I]) );
		pCommBlock->Steps[I].pFilterBuffer = NULL;
		pCommBlock->Steps[I].pMaskBuffer = NULL;
		pCommBlock->Steps[I].pInBuffer = NULL;
	}	
	
	//allocate buffers
	for( I = 0; I < StepCnt; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer = (unsigned char *)ExAllocatePool( NonPagedPool, pCommBlock->Steps[I].OutLen );
		if( pCommBlock->Steps[I].pOutBuffer == NULL )
		{
			AmUsbCommBlockFree( pCommBlock );
			return NULL;
		}
		if( pCommBlock->Steps[I].InLen > 0 )
		{
			pCommBlock->Steps[I].pFilterBuffer = (unsigned char *)ExAllocatePool( NonPagedPool, pCommBlock->Steps[I].InLen );
			pCommBlock->Steps[I].pMaskBuffer = (unsigned char *)ExAllocatePool( NonPagedPool, pCommBlock->Steps[I].InLen );
			pCommBlock->Steps[I].pInBuffer = (unsigned char *)ExAllocatePool( NonPagedPool, pCommBlock->Steps[I].InLen );
			if( pCommBlock->Steps[I].pFilterBuffer == NULL
			 || pCommBlock->Steps[I].pMaskBuffer == NULL
			 || pCommBlock->Steps[I].pInBuffer == NULL )
			{
				AmUsbCommBlockFree( pCommBlock );
				return NULL;
			}
		}
	} //for( I ...

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbCommBlockNew: new communication block %p\r\n", pCommBlock ) );

	//return pointer to communication block
	return pCommBlock;
}



//*** support functions for non-interrupt-time - device locked

//free device
//also frees URBs and their buffers
static void AmUsbFree( struct tAmUsbDev * pDev )
{
	LARGE_INTEGER Time;
	struct tAmUsbCommBlock * pCommBlock;

	AmUsbDebug( DBGLVL_MAXIMUM, ( "entering AmUsbFree\r\n" ) );

	//set unload flag (e.g. to prevent scheduling of DPCs)
	pDev->Unload = 1;

	//delete timeout timer
	if( pDev->TimeoutTimerActive )
	{
		KeCancelTimer( &pDev->TimeoutTimer );
		KeRemoveQueueDpc( &pDev->TimeoutDpc );
	}

	//remove DPC object from queue - unqueue tasks that cannot be done in interrupt
	KeRemoveQueueDpc( &pDev->DpcObject );

	//abort bulk input and output
	//(this frees all IRPs, URBs and buffers for bulk in and out)
	AmUsbOutCancel( pDev );
	AmUsbInCancel( pDev );

	//wait for bulk input / output to complete
	AmUsbDebug( DBGLVL_MAXIMUM, ( "AmUsbFree: starting to wait\r\n" ) );
	while( pDev->OutActive || pDev->InActive )
	{
		Time = RtlConvertLongToLargeInteger( -10000 ); //1ms relative time
		KeDelayExecutionThread( KernelMode, FALSE, &Time );
	}
	AmUsbDebug( DBGLVL_MAXIMUM, ( "AmUsbFree: waiting finshed\r\n" ) );

	//free communication blocks
	while( pDev->CommBlockList.pNext != &pDev->CommBlockList )
	{
		//remove first communication block from list
		pCommBlock = pDev->CommBlockList.pNext;
		pDev->CommBlockList.pNext = pDev->CommBlockList.pNext->pNext;
		pDev->CommBlockList.pNext->pPrev = &pDev->CommBlockList;
		//free communication block
		if( pCommBlock->FreeFlag )
			AmUsbCommBlockFree( pCommBlock );
	}

	//free device structure
	ExFreePool( pDev );

	AmUsbDebug( DBGLVL_MAXIMUM, ( "exiting AmUsbFree\r\n" ) );
}

//put communication block into communication block list
//(p_comm_block must be allocated with ExAllocatePool( NonPagedPool, ... ))
//(must be called with device locked)
static void AmUsbCommBlockAdd( struct tAmUsbDev * pDev, struct tAmUsbCommBlock * pCommBlock )
{
	//do nothing if already linked in list
	if( pCommBlock->Linked )
		return;

	//initialize communication block
	pCommBlock->OutStep = 0;
	pCommBlock->InStep = 0;
	pCommBlock->Timeout = 0;
	//put entry into list
	pCommBlock->pPrev = pDev->CommBlockList.pPrev;
	pCommBlock->pNext = &pDev->CommBlockList;
	pCommBlock->pNext->pPrev = pCommBlock;
	pCommBlock->pPrev->pNext = pCommBlock;
	pCommBlock->Linked = 1;
	
	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbCommBlockAdd: added communication block %p\r\n", pCommBlock ) );

	//set flag for execution of AmUsbCommBlockSend
	pDev->TaskCommBlockSend = 1;
	//schedule task for things that cannot be done in interrupt
	if( ! pDev->Unload )
		KeInsertQueueDpc( &pDev->DpcObject, NULL, NULL ); //no system arguments
}



//*** support functions for non-interrupt-time - device _not_ locked

//process a communication block
//returns STATUS_SUCCESS on success
//(may not be called with device locked)
static NTSTATUS AmUsbCommBlockDo( struct tAmUsbDev * pDev, struct tAmUsbCommBlock * pCommBlock )
{
	PRKEVENT pWaitEvent;
	KIRQL OldIrql;
	LARGE_INTEGER TimeoutTime;
	NTSTATUS Status;

	//allocate and initialize wait event to wait on
	pWaitEvent = (PRKEVENT)ExAllocatePool( NonPagedPool, sizeof( KEVENT ) );
	if( pWaitEvent == NULL )
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	KeInitializeEvent( pWaitEvent, NotificationEvent, FALSE );

	//initialize communication block: WaitEvent and FreeFlag
	pCommBlock->pWaitEvent = pWaitEvent;
	pCommBlock->FreeFlag = 0; //do not automatically free this communication block

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );
	//device was unplugged
	if( pDev->Removed )
	{
		//unlock private device data
		//(this restores the original IRQL afterwards)
		KeReleaseSpinLock( &pDev->Lock, OldIrql );
		//free wait event
		ExFreePool( pWaitEvent );
		//error
		return STATUS_DEVICE_DOES_NOT_EXIST; //ERROR_DEV_NOT_EXIST
	}

	//add communication block to list
	AmUsbCommBlockAdd( pDev, pCommBlock );

	//unlock private device data
	//(this restores the original IRQL afterwards)
	KeReleaseSpinLock( &pDev->Lock, OldIrql );

	//wait for communication block to be processed
	//I.e. wait on wait event
	TimeoutTime = RtlConvertLongToLargeInteger( pCommBlock->StepCnt * 2 * AmUsbTimeoutTime ); //steps * 2 * timeout time
	Status = KeWaitForSingleObject( pWaitEvent, UserRequest, KernelMode, TRUE, &TimeoutTime );

	//remove pointer to wait event from communication block
	pCommBlock->pWaitEvent = NULL;
	//free wait event
	ExFreePool( pWaitEvent );

	//check for errors
	if( Status != STATUS_SUCCESS
	 || pCommBlock->OutStep < pCommBlock->StepCnt
	 || pCommBlock->InStep < pCommBlock->StepCnt ) //communication block not completed
	{
		//timeout
		if( pCommBlock->Timeout || Status == STATUS_TIMEOUT )
			return STATUS_IO_TIMEOUT; //ERROR_SEM_TIMEOUT
		//error but not a timeout - interrupted by something
		return STATUS_DRIVER_INTERNAL_ERROR; //ERROR_IO_DEVICE
	}

	//success
	return STATUS_SUCCESS;
}



//*** file operations

//an userspace process is reading from one of the devices
NTSTATUS AmUsbFileRead( PFILE_OBJECT pFileObject, char * pBuffer, unsigned int Count, unsigned int * pLen )
{
	struct tAmUsbFilePriv * pFilePriv;
	struct tAmUsbDev * pDev;
	unsigned int Cnt;
	NTSTATUS Status;
	int I;
	unsigned long SecNo;
	struct tAmUsbCommBlock * pCommBlock;
	unsigned long Remainder;
	LARGE_INTEGER SecNoLarge;

	//default: no bytes read and success
	*pLen = 0;
	Cnt = 0;
	Status = STATUS_SUCCESS;

	//get private file data
	pFilePriv = (struct tAmUsbFilePriv *)pFileObject->FsContext;
	//no private file data
	if( pFilePriv == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE
	//get private device data
	pDev = pFilePriv->pDev;
	//no pointer to device data
	if( pDev == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//switch mode
	switch( pFilePriv->Mode )
	{
		//act as compact flash reader / writer
		case AmUsbModeCf:
			//buffer smaller than sector (512 bytes)
			if( Count < 512 )
			{
				Cnt = 0;
				break;
			}
			//get sector number to read
			SecNoLarge = RtlExtendedLargeIntegerDivide( pFileObject->CurrentByteOffset, 512, &Remainder );
			SecNo = SecNoLarge.LowPart;
			//read sector
			//allocate communication block
			pCommBlock = AmUsbCommBlockNew( 17, "\x05\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02",
												"\x06\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22" );
			if( pCommBlock == NULL )
			{
				Cnt = 0;
				Status = STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
				break;
			}
			//fill communication block
			pCommBlock->Steps[0].pOutBuffer[0] = 0x43;
			*(unsigned long *)(&pCommBlock->Steps[0].pOutBuffer[1]) = htonl( SecNo );
			memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x43\0\0\0\0\0\r\n", 6 );
			memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\r\n", 6 );
			for( I = 0; I < 16; I++ )
			{
				pCommBlock->Steps[I+1].pOutBuffer[0] = 0x31;
				pCommBlock->Steps[I+1].pOutBuffer[1] = I;
				pCommBlock->Steps[I+1].pFilterBuffer[0] = 0x31;
				pCommBlock->Steps[I+1].pFilterBuffer[1] = I;
				memset( &pCommBlock->Steps[I+1].pFilterBuffer[2], 0, 32 );
				pCommBlock->Steps[I+1].pMaskBuffer[0] = 0xFF;
				pCommBlock->Steps[I+1].pMaskBuffer[1] = 0xFF;
				memset( &pCommBlock->Steps[I+1].pMaskBuffer[2], 0, 32 );
			}
			//process commmunication block
			Status = AmUsbCommBlockDo( pDev, pCommBlock );
			if( Status != STATUS_SUCCESS )
			{
				Cnt = 0;
				AmUsbCommBlockFree( pCommBlock );
				break;
			}
			if( ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) ) != SecNo
			 || pCommBlock->Steps[0].pInBuffer[5] != 0 )
			{
				AmUsbCommBlockFree( pCommBlock );
				Cnt = 0;
				Status = STATUS_IO_DEVICE_ERROR; //ERROR_IO_DEVICE
				break;
			}
			//copy data to user space
			for( I = 0; I < 16; I++ )
				memcpy( &pBuffer[I << 5], &pCommBlock->Steps[I+1].pInBuffer[2], 32 );
			//free communication block
			AmUsbCommBlockFree( pCommBlock );
			//update file pointer
			SecNoLarge = RtlConvertUlongToLargeInteger( SecNo + 1 );
			pFileObject->CurrentByteOffset = RtlExtendedIntegerMultiply( SecNoLarge, 512 );
			//512 bytes read
			Cnt = 512;
			break; //case AmUsbModeCf

		//act as null device
		case AmUsbModeNull:
		//show stream written to device
		case AmUsbModeStream:
		default:
			//nothing to read
			Cnt = 0;
	} //switch( pFilePriv->Mode )

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbFileRead: read from file %08X: %d bytes\r\n", pFileObject, Cnt ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return number of bytes read and status
	*pLen = Cnt;
	return Status;
}

//an userspace process is writing to one of the devices
NTSTATUS AmUsbFileWrite( PFILE_OBJECT pFileObject, char * pBuffer, unsigned int Count, unsigned int * pLen )
{
	struct tAmUsbFilePriv * pFilePriv;
	struct tAmUsbDev * pDev;
	KIRQL OldIrql;
	unsigned int Cnt;
	NTSTATUS Status;
	int I, Pack, Ofs, Shift, Channels, MaxVal, X, Y, Xx, Yy;
	int FillL, FillR, FillT, FillB, SkipL, SkipT, StartPos;
	unsigned long Magic, SecNo;
	unsigned short Width, Height;
	unsigned char Val;
	struct tAmUsbCommBlock * pCommBlock;
	unsigned long Remainder;
	LARGE_INTEGER SecNoLarge;

	//default: no bytes written and success
	*pLen = 0;
	Cnt = 0;
	Status = STATUS_SUCCESS;

	//get private file data
	pFilePriv = (struct tAmUsbFilePriv *)pFileObject->FsContext;
	//no private file data
	if( pFilePriv == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE
	//get private device data
	pDev = pFilePriv->pDev;
	//no pointer to device data
	if( pDev == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//switch mode
	switch( pFilePriv->Mode )
	{
		//show stream written to device
		case AmUsbModeStream:
			//default: only one byte written
			// - idea is to skip only the first byte if the data could not be understood
			if( Count >= 1 )
				Cnt = 1;
			else
				Cnt = 0;

			//get magic
			Magic = 0;
			if( Count >= 4 )
			{
				//get magic
				Magic = ntohl( *(unsigned long *)pBuffer );
			}
			//get parameters
			switch( Magic )
			{
				//DEADBEEF and FEEDBEEF frame
				case 0xDEADBEEF:
				case 0xFEEDBEEF:
					//check length of packet
					if( Count < 12 )
					{
						Magic = 0;
						Width = 0;
						Height = 0;
						Channels = 1;
						MaxVal = 1;
						StartPos = 4;
						break;
					}
					//get width and height
					Width = ntohs( *(unsigned short *)(pBuffer+8) );
					Height = ntohs( *(unsigned short *)(pBuffer+10) );
					//set rest of parameters
					Channels = 1;
					MaxVal = Magic == 0xFEEDBEEF ? 0xFF : 1;
					StartPos = 12;
					break;
				//MCU frame
				case 0x23542666:
					//check length of packet
					if( Count < 12 )
					{
						Magic = 0;
						Width = 0;
						Height = 0;
						Channels = 1;
						MaxVal = 1;
						StartPos = 4;
						break;
					}
					//get height and widtht
					Height = ntohs( *(unsigned short *)(pBuffer+4) );
					Width = ntohs( *(unsigned short *)(pBuffer+6) );
					//get channels and maxval
					Channels = ntohs( *(unsigned short *)(pBuffer+8) );
					MaxVal =  ntohs( *(unsigned short *)(pBuffer+10) );
					//set rest of parameters
					StartPos = 12;
					break;
				default:
					Magic = 0;
					Width = 0;
					Height = 0;
					Channels = 1;
					MaxVal = 1;
					StartPos = 4;
					break;
			}
			//some valid packet
			while( Magic != 0 ) //use while command instead of if command to have break command
			{
				//check length of packet
				if( Count < (unsigned int)StartPos + (unsigned int)Width * (unsigned int)Height * (unsigned int)Channels )
					{ Magic = 0; break; }
				//calculate parameters
				FillL = Width < 26 ? (26 - Width) / 2 : 0;
				FillR = Width < 26 ? 26 - Width - FillL : 0;
				FillT = Height < 20 ? (20 - Height) / 2 : 0;
				FillB = Height < 20 ? 20 - Height - FillT : 0;
				SkipL = Width > 26 ? (Width - 26) / 2 : 0;
				SkipT = Height > 20 ? (Height - 20) / 2 : 0;
				//allocate communication block
				pCommBlock = AmUsbCommBlockNew( 6, "\x36\x36\x36\x36\x36\x01\r\n", "\0\0\0\0\0\0" );
				if( pCommBlock == NULL )
					break;
				//fill communication block
				for( Pack = 0; Pack < 5; Pack++ )
				{
					pCommBlock->Steps[Pack].pOutBuffer[0] = 0x10;
					pCommBlock->Steps[Pack].pOutBuffer[1] = (unsigned char)Pack;
					for( I = 2; I < 54; I++ )
						pCommBlock->Steps[Pack].pOutBuffer[I] = 0x00;
				}
				for( Y = 0; Y < 20; Y++ )
				for( X = 0; X < 26; X++ )
				{
					//get packet and offset
					I = Y * 13 + X / 2;
					Pack = I / 52;
					Ofs = I % 52 + 2;
					Shift = (1 - X % 2) * 4;
					//real pixel
					if( X >= FillL && X < 26 - FillR && Y >= FillT && Y < 20 - FillB )
					{
						Xx = X - FillL + SkipL;
						Yy = Y - FillT + SkipT;
						Val = *(pBuffer + StartPos + (Yy * Width + Xx) * Channels);
						Val = Val > MaxVal ? 0x0F : (unsigned char)((unsigned long)(0x0F * Val) / MaxVal);
						pCommBlock->Steps[Pack].pOutBuffer[Ofs] |= Val << Shift;
					}
				}
				//fill show packet with data
				pCommBlock->Steps[5].pOutBuffer[0] = 0x20;
				//lock private device data
				//(this first raises IRQL to DISPATCH_LEVEL)
				KeAcquireSpinLock( &pDev->Lock, &OldIrql );
				//device was unplugged
				if( pDev->Removed )
				{
					//free communication block
					AmUsbCommBlockFree( pCommBlock );
					//unlock private device data
					//(this restores the original IRQL afterwards)
					KeReleaseSpinLock( &pDev->Lock, OldIrql );
					//unlock device for file operations
					ExReleaseFastMutex( &pDev->FileLock );
					//error
					return STATUS_DEVICE_DOES_NOT_EXIST; //ERROR_DEV_NOT_EXIST
				}
				//submit the communication block
				pCommBlock->pWaitEvent = NULL; //do not signal completion
				pCommBlock->FreeFlag = 1; //automatically free after processing
				AmUsbCommBlockAdd( pDev, pCommBlock );
				//unlock private device data
				//(this restores the original IRQL afterwards)
				KeReleaseSpinLock( &pDev->Lock, OldIrql );
				//return number of bytes written to this device
				Cnt = StartPos + (unsigned int)Width * (unsigned int)Height * (unsigned int)Channels;
				//we used the while command only to have the break command
				//we wanted an if command
				break;
			}
			break; //case AmUsbModeStream

		//act as compact flash reader / writer
		case AmUsbModeCf:
			//no data to write
			if( Count == 0 )
			{
				Cnt = 0;
				break;
			}
			//get sector number to write
			SecNoLarge = RtlExtendedLargeIntegerDivide( pFileObject->CurrentByteOffset, 512, &Remainder );
			SecNo = SecNoLarge.LowPart;
			//write sector
			//allocate communication block
			pCommBlock = AmUsbCommBlockNew( 17, "\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x05",
												"\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x06" );
			if( pCommBlock == NULL )
			{
				Cnt = 0;
				Status = STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
				break;
			}
			//fill communication block
			for( I = 0; I < 16; I++ )
			{
				pCommBlock->Steps[I].pOutBuffer[0] = 0x30;
				pCommBlock->Steps[I].pOutBuffer[1] = I;
				memset( &pCommBlock->Steps[I].pOutBuffer[2], 0, 32 );
				if( (int)Count >= (I << 5) + 32 )
					memcpy( &pCommBlock->Steps[I].pOutBuffer[2], &pBuffer[I << 5], 32 );
				else if( (int)Count > I << 5 )
					memcpy( &pCommBlock->Steps[I].pOutBuffer[2], &pBuffer[I << 5], Count - (I << 5) );
				pCommBlock->Steps[I].pFilterBuffer[0] = 0x30;
				pCommBlock->Steps[I].pFilterBuffer[1] = I;
				pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
				pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
			}
			pCommBlock->Steps[16].pOutBuffer[0] = 0x42;
			*(unsigned long *)(&pCommBlock->Steps[16].pOutBuffer[1]) = htonl( SecNo );
			memcpy( pCommBlock->Steps[16].pFilterBuffer, "\x42\0\0\0\0\0\r\n", 6 );
			memcpy( pCommBlock->Steps[16].pMaskBuffer, "\xFF\0\0\0\0\0\r\n", 6 );
			//process commmunication block
			Status = AmUsbCommBlockDo( pDev, pCommBlock );
			if( Status != STATUS_SUCCESS )
			{
				Cnt = 0;
				AmUsbCommBlockFree( pCommBlock );
				break;
			}
			if( ntohl( *(unsigned long *)(&pCommBlock->Steps[16].pInBuffer[1]) ) != SecNo
			 || pCommBlock->Steps[16].pInBuffer[5] != 0 )
			{
				AmUsbCommBlockFree( pCommBlock );
				Cnt = 0;
				Status = STATUS_IO_DEVICE_ERROR; //ERROR_IO_DEVICE
				break;
			}
			//free communication block
			AmUsbCommBlockFree( pCommBlock );
			//update file pointer
			SecNoLarge = RtlConvertUlongToLargeInteger( SecNo + 1 );
			pFileObject->CurrentByteOffset = RtlExtendedIntegerMultiply( SecNoLarge, 512 );
			//512 bytes read (if 512 bytes were supplied)
			Cnt = min( (int)Count, 512 );
			break; //case AmUsbModeCf

		//act as null device
		case AmUsbModeNull:
		default:
			//write everything
			Cnt = Count;
	} //switch( pFilePriv->Mode )

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbFileWrite: wrote to file %08X: %d bytes\r\n", pFileObject, Cnt ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return number of bytes written and status
	*pLen = Cnt;
	return Status;
}

//halt ARCADEmini
//turns off command processing
static NTSTATUS AmUsbFileIoctlHalt( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x01" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x00;
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x00;
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl Halt on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//run ARCADEmini
//turns on command processing
static NTSTATUS AmUsbFileIoctlRun( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x01" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x01;
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x01;
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl Run on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//step ARCADEmini
//processes a single command
static NTSTATUS AmUsbFileIoctlStep( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x01" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x02;
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x02;
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl Step on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//set command pointer
static NTSTATUS AmUsbFileIoctlSetCmdPtr( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbSetCmdPtrIn * pSetCmdPtrIn, struct tAmUsbSetCmdPtrOut * pSetCmdPtrOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x07", "\x07" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x08;
	*(unsigned long *)(&pCommBlock->Steps[0].pOutBuffer[1]) = htonl( pSetCmdPtrIn->SecNo );
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[5]) = htons( pSetCmdPtrIn->Ofs );
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x08;
	memset( &pCommBlock->Steps[0].pFilterBuffer[1], 0, 6 );
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	memset( &pCommBlock->Steps[0].pMaskBuffer[1], 0, 6 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pSetCmdPtrOut->SecNo = ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) );
		pSetCmdPtrOut->Ofs = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[5]) );
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SetCmdPtr on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get command pointer
static NTSTATUS AmUsbFileIoctlGetCmdPtr( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetCmdPtrOut * pGetCmdPtrOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x07" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x09;
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x09;
	memset( &pCommBlock->Steps[0].pFilterBuffer[1], 0, 6 );
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	memset( &pCommBlock->Steps[0].pMaskBuffer[1], 0, 6 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pGetCmdPtrOut->SecNo = ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) );
		pGetCmdPtrOut->Ofs = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[5]) );
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetCmdPtr on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//load pixel buffer
//loads the pixel buffer with 260 bytes of data
static NTSTATUS AmUsbFileIoctlLoadPixBuf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbLoadPixBufIn * pLoadPixBufIn )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 5, "\x36\x36\x36\x36\x36", "\x02\x02\x02\x02\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 5; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x10;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		memcpy( &pCommBlock->Steps[I].pOutBuffer[2], &pLoadPixBufIn->Pixels[I * 52], 52 );
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x10;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl LoadPixBuf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//read pixel buffer
//reads 260 bytes of data from the pixel buffer
static NTSTATUS AmUsbFileIoctlReadPixBuf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbReadPixBufOut * pReadPixBufOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 5, "\x02\x02\x02\x02\x02", "\x36\x36\x36\x36\x36" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 5; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x11;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x11;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		memset( &pCommBlock->Steps[I].pFilterBuffer[2], 0, 52 );
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
		memset( &pCommBlock->Steps[I].pMaskBuffer[2], 0, 52 );
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//copy data to user
		for( I = 0; I < 5; I++ )
			memcpy( &pReadPixBufOut->Pixels[I * 52], &pCommBlock->Steps[I].pInBuffer[2], 52 );
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl ReadPixBuf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//load overlay pixel buffer
//loads the overlay pixel buffer with 260 bytes of data
static NTSTATUS AmUsbFileIoctlLoadOvpBuf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbLoadOvpBufIn * pLoadOvpBufIn )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 5, "\x36\x36\x36\x36\x36", "\x02\x02\x02\x02\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 5; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x12;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		memcpy( &pCommBlock->Steps[I].pOutBuffer[2], &pLoadOvpBufIn->Pixels[I * 52], 52 );
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x12;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl LoadOvpBuf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//read overlay pixel buffer
//reads 260 bytes of data from the overlay pixel buffer
static NTSTATUS AmUsbFileIoctlReadOvpBuf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbReadOvpBufOut * pReadOvpBufOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 5, "\x02\x02\x02\x02\x02", "\x36\x36\x36\x36\x36" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 5; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x13;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x13;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		memset( &pCommBlock->Steps[I].pFilterBuffer[2], 0, 52 );
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
		memset( &pCommBlock->Steps[I].pMaskBuffer[2], 0, 52 );
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//copy data to user
		for( I = 0; I < 5; I++ )
			memcpy( &pReadOvpBufOut->Pixels[I * 52], &pCommBlock->Steps[I].pInBuffer[2], 52 );
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl ReadOvpBuf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//swap display buffers
//exchanges the foreground and background pixel buffer
static NTSTATUS AmUsbFileIoctlSwapBufs( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x01" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x20;
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x20;
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SwapBufs on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//swap display buffers and copy image
//exchanges the foreground and background pixel buffer
//then copies the foreground picture into the background buffer
static NTSTATUS AmUsbFileIoctlSwapBufsCp( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x01" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x21;
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x21;
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SwapBufsCp on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//load sector buffer
//loads the sector buffer with 512 bytes of data
static NTSTATUS AmUsbFileIoctlLoadSecBuf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbLoadSecBufIn * pLoadSecBufIn )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 16, "\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22",
										"\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 16; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x30;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		memcpy( &pCommBlock->Steps[I].pOutBuffer[2], &pLoadSecBufIn->Data[I << 5], 32 );
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x30;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl LoadSecBuf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//read sector buffer
//reads 512 bytes of data from the sector buffer
static NTSTATUS AmUsbFileIoctlReadSecBuf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbReadSecBufOut * pReadSecBufOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 16, "\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02\x02",
										"\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22\x22" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 16; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x31;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x31;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		memset( &pCommBlock->Steps[I].pFilterBuffer[2], 0, 32 );
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
		memset( &pCommBlock->Steps[I].pMaskBuffer[2], 0, 32 );
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//copy data to user
		for( I = 0; I < 16; I++ )
			memcpy( &pReadSecBufOut->Data[I << 5], &pCommBlock->Steps[I].pInBuffer[2], 32 );
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl ReadSecBuf on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get CF sector count
//returns the number of sectors on compact flash
static NTSTATUS AmUsbFileIoctlGetCfSecCnt( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetCfSecCntOut * pGetCfSecCntOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x05" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x40\r\n", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x40\0\0\0\0\r\n", 5 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\r\n", 5 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
		//return data to user
		pGetCfSecCntOut->SecCnt = ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetCfSecCnt on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//write CF sector
//writes the sector buffer to a sector on compact flash
static NTSTATUS AmUsbFileIoctlWriteCf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbWriteCfIn * pWriteCfIn, struct tAmUsbWriteCfOut * pWriteCfOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x05", "\x06" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x42;
	*(unsigned long *)(&pCommBlock->Steps[0].pOutBuffer[1]) = htonl( pWriteCfIn->SecNo );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x42\0\0\0\0\0\r\n", 6 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\r\n", 6 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pWriteCfOut->SecNo = ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) );
		pWriteCfOut->Err = pCommBlock->Steps[0].pInBuffer[5];
	}
	
	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl WriteCf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//read CF sector
//reads a sector on compact flash into sector buffer
static NTSTATUS AmUsbFileIoctlReadCf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbReadCfIn * pReadCfIn, struct tAmUsbReadCfOut * pReadCfOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x05", "\x06" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x43;
	*(unsigned long *)(&pCommBlock->Steps[0].pOutBuffer[1]) = htonl( pReadCfIn->SecNo );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x43\0\0\0\0\0\r\n", 6 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\r\n", 6 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pReadCfOut->SecNo = ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) );
		pReadCfOut->Err = pCommBlock->Steps[0].pInBuffer[5];
	}
	
	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl ReadCf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get internal sector count
//returns the number of sectors in internal flash
static NTSTATUS AmUsbFileIoctlGetIfSecCnt( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetIfSecCntOut * pGetIfSecCntOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x05" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x44\r\n", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x44\0\0\0\0\r\n", 5 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\r\n", 5 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
		//return data to user
		pGetIfSecCntOut->SecCnt = ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetIfSecCnt on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//read internal sector
//reads a sector in internal flash into sector buffer
static NTSTATUS AmUsbFileIoctlReadIf( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbReadIfIn * pReadIfIn, struct tAmUsbReadIfOut * pReadIfOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x05", "\x06" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x47;
	*(unsigned long *)(&pCommBlock->Steps[0].pOutBuffer[1]) = htonl( pReadIfIn->SecNo );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x47\0\0\0\0\0\r\n", 6 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\r\n", 6 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pReadIfOut->SecNo = ntohl( *(unsigned long *)(&pCommBlock->Steps[0].pInBuffer[1]) );
		pReadIfOut->Err = pCommBlock->Steps[0].pInBuffer[5];
	}

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl ReadIf on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get compact flash flag
//returns if commands are executed from compact flash or from internal flash
static NTSTATUS AmUsbFileIoctlGetCfFlag( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetCfFlagOut * pGetCfFlagOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x48\r\n", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x48\0\r\n", 2 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\r\n", 2 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
		//return data to user
		pGetCfFlagOut->CfFlag = pCommBlock->Steps[0].pInBuffer[1] == 0 ? 0 : 1;

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetCfFlag on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//load variables
//loads the variables with 256 bytes of data
static NTSTATUS AmUsbFileIoctlLoadVars( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbLoadVarsIn * pLoadVarsIn )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 8, "\x22\x22\x22\x22\x22\x22\x22\x22",
									   "\x02\x02\x02\x02\x02\x02\x02\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 8; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x50;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		memcpy( &pCommBlock->Steps[I].pOutBuffer[2], &pLoadVarsIn->Vars[I << 5], 32 );
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x50;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl LoadVars on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//read variables
//reads 256 bytes of data from the variables
static NTSTATUS AmUsbFileIoctlReadVars( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbReadVarsOut * pReadVarsOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 8, "\x02\x02\x02\x02\x02\x02\x02\x02",
									   "\x22\x22\x22\x22\x22\x22\x22\x22" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 8; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x51;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x51;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		memset( &pCommBlock->Steps[I].pFilterBuffer[2], 0, 32 );
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
		memset( &pCommBlock->Steps[I].pMaskBuffer[2], 0, 32 );
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//copy data to user
		for( I = 0; I < 8; I++ )
			memcpy( &pReadVarsOut->Vars[I << 5], &pCommBlock->Steps[I].pInBuffer[2], 32 );
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl ReadVars on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//load stack
//loads the stack with 128 bytes of data
static NTSTATUS AmUsbFileIoctlLoadStack( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbLoadStackIn * pLoadStackIn )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 4, "\x22\x22\x22\x22", "\x02\x02\x02\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 4; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x54;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		memcpy( &pCommBlock->Steps[I].pOutBuffer[2], &pLoadStackIn->Data[I << 5], 32 );
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x54;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl LoadStack on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//read stack
//reads 128 bytes of data from the stack
static NTSTATUS AmUsbFileIoctlReadStack( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbReadStackOut * pReadStackOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	int I, Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 4, "\x02\x02\x02\x02", "\x22\x22\x22\x22" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	for( I = 0; I < 4; I++ )
	{
		pCommBlock->Steps[I].pOutBuffer[0] = 0x55;
		pCommBlock->Steps[I].pOutBuffer[1] = I;
		pCommBlock->Steps[I].pFilterBuffer[0] = 0x55;
		pCommBlock->Steps[I].pFilterBuffer[1] = I;
		memset( &pCommBlock->Steps[I].pFilterBuffer[2], 0, 32 );
		pCommBlock->Steps[I].pMaskBuffer[0] = 0xFF;
		pCommBlock->Steps[I].pMaskBuffer[1] = 0xFF;
		memset( &pCommBlock->Steps[I].pMaskBuffer[2], 0, 32 );
	}
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//copy data to user
		for( I = 0; I < 4; I++ )
			memcpy( &pReadStackOut->Data[I << 5], &pCommBlock->Steps[I].pInBuffer[2], 32 );
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl ReadStack on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//set stack pointer
static NTSTATUS AmUsbFileIoctlSetStackPtr( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbSetStackPtrIn * pSetStackPtrIn, struct tAmUsbSetStackPtrOut * pSetStackPtrOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x02", "\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x56;
	pCommBlock->Steps[0].pOutBuffer[1] = pSetStackPtrIn->StackPtr;
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x56\0\r\n", 2 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\r\n", 2 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
		//return data to user
		pSetStackPtrOut->StackPtr = pCommBlock->Steps[0].pInBuffer[1];

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SetStackPtr on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get stack pointer
static NTSTATUS AmUsbFileIoctlGetStackPtr( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetStackPtrOut * pGetStackPtrOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x57\r\n", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x57\0\r\n", 2 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\r\n", 2 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
		//return data to user
		pGetStackPtrOut->StackPtr = pCommBlock->Steps[0].pInBuffer[1];

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetStackPtr on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get and set key state
//gets key state and sets key state afterwards
static NTSTATUS AmUsbFileIoctlGetSetKeys( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetSetKeysIn * pGetSetKeysIn, struct tAmUsbGetSetKeysOut * pGetSetKeysOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x05", "\x04" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x60;
	pCommBlock->Steps[0].pOutBuffer[1] = pGetSetKeysIn->Pressed;
	pCommBlock->Steps[0].pOutBuffer[2] = pGetSetKeysIn->PressedMask;
	pCommBlock->Steps[0].pOutBuffer[3] = pGetSetKeysIn->Released;
	pCommBlock->Steps[0].pOutBuffer[4] = pGetSetKeysIn->ReleasedMask;
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x60\0\0\0\r\n", 4 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\r\n", 4 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pGetSetKeysOut->State = pCommBlock->Steps[0].pInBuffer[1];
		pGetSetKeysOut->Pressed = pCommBlock->Steps[0].pInBuffer[2];
		pGetSetKeysOut->Released = pCommBlock->Steps[0].pInBuffer[3];
	}

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetSetKeys on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get and set SNES0 key state
//gets SNES0 key state and sets SNES0 key state afterwards
static NTSTATUS AmUsbFileIoctlGetSetSnes0( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetSetSnes0In * pGetSetSnes0In, struct tAmUsbGetSetSnes0Out * pGetSetSnes0Out )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x09", "\x07" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x62;
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[1]) = htons( pGetSetSnes0In->Pressed );
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[3]) = htons( pGetSetSnes0In->PressedMask );
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[5]) = htons( pGetSetSnes0In->Released );
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[7]) = htons( pGetSetSnes0In->ReleasedMask );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x62\0\0\0\0\0\0\r\n", 7 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\0\r\n", 7 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pGetSetSnes0Out->State = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[1]) );
		pGetSetSnes0Out->Pressed = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[3]) );
		pGetSetSnes0Out->Released = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[5]) );
	}
	
	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetSetSnes0 on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get and set SNES1 key state
//gets SNES1 key state and sets SNES1 key state afterwards
static NTSTATUS AmUsbFileIoctlGetSetSnes1( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetSetSnes1In * pGetSetSnes1In, struct tAmUsbGetSetSnes1Out * pGetSetSnes1Out )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x09", "\x07" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x63;
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[1]) = htons( pGetSetSnes1In->Pressed );
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[3]) = htons( pGetSetSnes1In->PressedMask );
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[5]) = htons( pGetSetSnes1In->Released );
	*(unsigned short *)(&pCommBlock->Steps[0].pOutBuffer[7]) = htons( pGetSetSnes1In->ReleasedMask );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x63\0\0\0\0\0\0", 7 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\0", 7 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pGetSetSnes1Out->State = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[1]) );
		pGetSetSnes1Out->Pressed = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[3]) );
		pGetSetSnes1Out->Released = ntohs( *(unsigned short *)(&pCommBlock->Steps[0].pInBuffer[5]) );
	}
	
	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetSetSnes1 on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//set time
//sets the time of the ARCADEmini closk
static NTSTATUS AmUsbFileIoctlSetTime( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbSetTimeIn * pSetTimeIn )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x04", "\x01" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x70;
	pCommBlock->Steps[0].pOutBuffer[1] = pSetTimeIn->Sec;
	pCommBlock->Steps[0].pOutBuffer[2] = pSetTimeIn->Min;
	pCommBlock->Steps[0].pOutBuffer[3] = pSetTimeIn->Hour;
	pCommBlock->Steps[0].pFilterBuffer[0] = 0x70;
	pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SetTime on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get time
//gets the time from the ARCADEmini clock
static NTSTATUS AmUsbFileIoctlGetTime( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetTimeOut * pGetTimeOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x04" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x71", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x71\0\0\0", 4 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0", 4 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pGetTimeOut->Sec = pCommBlock->Steps[0].pInBuffer[1];
		pGetTimeOut->Min = pCommBlock->Steps[0].pInBuffer[2];
		pGetTimeOut->Hour = pCommBlock->Steps[0].pInBuffer[3];
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetTime on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get firmware version
//returns the firmware version of ARCADEmini
static NTSTATUS AmUsbFileIoctlGetFwVersion( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetFwVersionOut * pGetFwVersionOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x07" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x80", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x80\0\0\0\0\0\0", 7 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\0", 7 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( NT_SUCCESS( Status ) )
	{
		//return data to user
		pGetFwVersionOut->Major = pCommBlock->Steps[0].pInBuffer[1];
		pGetFwVersionOut->Minor = pCommBlock->Steps[0].pInBuffer[2];
		pGetFwVersionOut->Revision = pCommBlock->Steps[0].pInBuffer[3];
		pGetFwVersionOut->Year = pCommBlock->Steps[0].pInBuffer[4];
		pGetFwVersionOut->Month = pCommBlock->Steps[0].pInBuffer[5];
		pGetFwVersionOut->Day = pCommBlock->Steps[0].pInBuffer[6];
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetFwVersion on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get ID
//gets the ID of this ARCADEmini
static NTSTATUS AmUsbFileIoctlGetId( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetIdOut * pGetIdOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x12" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x84", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x84\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 18 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 18 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		memcpy( pGetIdOut->Id, &pCommBlock->Steps[0].pInBuffer[1], 16 );
		pGetIdOut->Err = pCommBlock->Steps[0].pInBuffer[17];
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetId on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//set ID
//sets the ID of this ARCADEmini
static NTSTATUS AmUsbFileIoctlSetId( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbSetIdIn * pSetIdIn, struct tAmUsbSetIdOut * pSetIdOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x11", "\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x85;
	memcpy( &pCommBlock->Steps[0].pOutBuffer[1], pSetIdIn->Id, 16 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x85\0", 2 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0", 2 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pSetIdOut->Err = pCommBlock->Steps[0].pInBuffer[1];
	}

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SetId on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get output styling
//gets the output styling
static NTSTATUS AmUsbFileIoctlGetOutStyle( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetOutStyleOut * pGetOutStyleOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x03" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x88", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x88\0\0", 3 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0", 3 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pGetOutStyleOut->Style = pCommBlock->Steps[0].pInBuffer[1];
		pGetOutStyleOut->Err = pCommBlock->Steps[0].pInBuffer[2];
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetOutStyle on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//set output style
//sets the output style
static NTSTATUS AmUsbFileIoctlSetOutStyle( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbSetOutStyleIn * pSetOutStyleIn, struct tAmUsbSetOutStyleOut * pSetOutStyleOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x02", "\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x89;
	pCommBlock->Steps[0].pOutBuffer[1] = pSetOutStyleIn->Style;
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x89\0", 2 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0", 2 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pSetOutStyleOut->Err = pCommBlock->Steps[0].pInBuffer[1];
	}

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SetOutStyle on file %p: %08X\r\n", pFileObject, Status ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get output timing
//gets the output timing
static NTSTATUS AmUsbFileIoctlGetOutTiming( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetOutTimingOut * pGetOutTimingOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x11" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	memcpy( pCommBlock->Steps[0].pOutBuffer, "\x8A", 1 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x8A\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 17 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0", 18 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		memcpy( pGetOutTimingOut->Timing, &pCommBlock->Steps[0].pInBuffer[1], 15 );
		pGetOutTimingOut->Err = pCommBlock->Steps[0].pInBuffer[16];
	}

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetOutTiming on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//set output timing
//sets the output timing
static NTSTATUS AmUsbFileIoctlSetOutTiming( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbSetOutTimingIn * pSetOutTimingIn, struct tAmUsbSetOutTimingOut * pSetOutTimingOut )
{
	struct tAmUsbCommBlock * pCommBlock;
	NTSTATUS Status;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//allocate communication block
	pCommBlock = AmUsbCommBlockNew( 1, "\x10", "\x02" );
	if( pCommBlock == NULL )
	{
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}
	//fill communication block
	pCommBlock->Steps[0].pOutBuffer[0] = 0x8B;
	memcpy( &pCommBlock->Steps[0].pOutBuffer[1], pSetOutTimingIn->Timing, 15 );
	memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x8B\0", 2 );
	memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0", 2 );
	//process commmunication block
	Status = AmUsbCommBlockDo( pDev, pCommBlock );

	//free communication block
	AmUsbCommBlockFree( pCommBlock );

	//success
	if( Status == 0 )
	{
		//return data to user
		pSetOutTimingOut->Err = pCommBlock->Steps[0].pInBuffer[1];
	}

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SetOutTiming on file %p: %08X\r\n", pFileObject, Status ) );
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return with error code
	return Status;
}

//get version information
//returns version information about the driver
static NTSTATUS AmUsbFileIoctlGetVersion( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetVersionOut * pGetVersionOut )
{
	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//return data to user
	pGetVersionOut->Major = AmUsbVerMajor;
	pGetVersionOut->Minor = AmUsbVerMinor;
	pGetVersionOut->Revision = AmUsbVerRevision;
	pGetVersionOut->Year = AmUsbVerYear;
	pGetVersionOut->Month = AmUsbVerMonth;
	pGetVersionOut->Day = AmUsbVerDay;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetVersion on file %p\r\n", pFileObject ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return success
	return STATUS_SUCCESS;
}

//set mode
//sets the driver mode for normal file operations
static NTSTATUS AmUsbFileIoctlSetMode( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbSetModeIn * pSetModeIn, struct tAmUsbSetModeOut * pSetModeOut )
{
	KIRQL OldIrql;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );
	//get new mode
	switch( pSetModeIn->Mode )
	{
		//known modes
		case AmUsbModeNull:
		case AmUsbModeStream:
		case AmUsbModeCf:
			//accept new mode for this file
			pFilePriv->Mode = pSetModeIn->Mode;
			break;
	}
	//get new default mode from local structure
	switch( pSetModeIn->DefMode )
	{
		//known modes
		case AmUsbModeNull:
		case AmUsbModeStream:
		case AmUsbModeCf:
			//accept new default mode for this device
			pDev->Mode = pSetModeIn->DefMode;
			break;
	}
	//return data to user
	pSetModeOut->Mode = pFilePriv->Mode;
	pSetModeOut->DefMode = pDev->Mode;
	//unlock private device data
	//(this restores the original IRQL afterwards)
	KeReleaseSpinLock( &pDev->Lock, OldIrql );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl SetMode on file %p\r\n", pFileObject ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return success
	return STATUS_SUCCESS;
}

//get mode
//gets the driver mode for normal file operations
static NTSTATUS AmUsbFileIoctlGetMode( PFILE_OBJECT pFileObject, struct tAmUsbFilePriv * pFilePriv, struct tAmUsbDev * pDev, struct tAmUsbGetModeOut * pGetModeOut )
{
	KIRQL OldIrql;

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );
	//return data to user
	pGetModeOut->Mode = pFilePriv->Mode;
	pGetModeOut->DefMode = pDev->Mode;
	//unlock private device data
	//(this restores the original IRQL afterwards)
	KeReleaseSpinLock( &pDev->Lock, OldIrql );

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "ioctl GetMode on file %p\r\n", pFileObject ) );

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//return success
	return STATUS_SUCCESS;
}

//an userspace process is doing an ioctl on one of the devices
NTSTATUS AmUsbFileIoctl( PFILE_OBJECT pFileObject, unsigned int IoctlNo, void * pBuffer, unsigned int InLen, unsigned int OutBufLen, unsigned int * pOutLen )
{
	struct tAmUsbFilePriv * pFilePriv;
	struct tAmUsbDev * pDev;

	//default: no bytes returned
	*pOutLen = 0;

	//get private file data
	pFilePriv = (struct tAmUsbFilePriv *)pFileObject->FsContext;
	//no private file data
	if( pFilePriv == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE
	//get private device data
	pDev = pFilePriv->pDev;
	//no pointer to device data
	if( pDev == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE

	//choose ioctl
	switch( IoctlNo )
	{
//macros to simplify this switch
#define AmUsbFileIoctlSwitchCase( IoctlNo, IoctlFunc ) \
	case IoctlNo: \
		*pOutLen = 0; \
		return IoctlFunc( pFileObject, pFilePriv, pDev );
#define AmUsbFileIoctlSwitchCaseIn( IoctlNo, IoctlFunc, InType ) \
	case IoctlNo: \
		*pOutLen = 0; \
		if( InLen < sizeof( struct InType ) ) /*input buffer too small*/ \
			return STATUS_BUFFER_TOO_SMALL; /*ERROR_INSUFFICIENT_BUFFER*/ \
		return IoctlFunc( pFileObject, pFilePriv, pDev, (struct InType *)pBuffer );
#define AmUsbFileIoctlSwitchCaseOut( IoctlNo, IoctlFunc, OutType ) \
	case IoctlNo: \
		if( OutBufLen < sizeof( struct OutType ) ) /*output buffer too small*/ { \
			*pOutLen = 0; \
			return STATUS_BUFFER_TOO_SMALL; /*ERROR_INSUFFICIENT_BUFFER*/ } \
		*pOutLen = sizeof( struct OutType ); \
		return IoctlFunc( pFileObject, pFilePriv, pDev, (struct OutType *)pBuffer );
#define AmUsbFileIoctlSwitchCaseInOut( IoctlNo, IoctlFunc, InType, OutType ) \
	case IoctlNo: \
		if( InLen < sizeof( struct InType ) /*input buffer too small*/ \
		 || OutBufLen < sizeof( struct OutType ) ) /*output buffer too small*/ { \
			*pOutLen = 0; \
			return STATUS_BUFFER_TOO_SMALL; /*ERROR_INSUFFICIENT_BUFFER*/ } \
		*pOutLen = sizeof( struct OutType ); \
		return IoctlFunc( pFileObject, pFilePriv, pDev, (struct InType *)pBuffer, (struct OutType *)pBuffer );

		//halt ARCADEmini
		AmUsbFileIoctlSwitchCase( AmUsbIoctlNoHalt, AmUsbFileIoctlHalt );
		//run ARCADEmini
		AmUsbFileIoctlSwitchCase( AmUsbIoctlNoRun, AmUsbFileIoctlRun );
		//step ARCADEmini
		AmUsbFileIoctlSwitchCase( AmUsbIoctlNoStep, AmUsbFileIoctlStep );
		//set command pointer
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoSetCmdPtr, AmUsbFileIoctlSetCmdPtr, tAmUsbSetCmdPtrIn, tAmUsbSetCmdPtrOut );
		//get command pointer
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetCmdPtr, AmUsbFileIoctlGetCmdPtr, tAmUsbGetCmdPtrOut );
		//load pixel buffer
		AmUsbFileIoctlSwitchCaseIn( AmUsbIoctlNoLoadPixBuf, AmUsbFileIoctlLoadPixBuf, tAmUsbLoadPixBufIn );
		//read pixel buffer
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoReadPixBuf, AmUsbFileIoctlReadPixBuf, tAmUsbReadPixBufOut );
		//load overlay pixel buffer
		AmUsbFileIoctlSwitchCaseIn( AmUsbIoctlNoLoadOvpBuf, AmUsbFileIoctlLoadOvpBuf, tAmUsbLoadOvpBufIn );
		//read overlay pixel buffer
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoReadOvpBuf, AmUsbFileIoctlReadOvpBuf, tAmUsbReadOvpBufOut );
		//swap display buffers
		AmUsbFileIoctlSwitchCase( AmUsbIoctlNoSwapBufs, AmUsbFileIoctlSwapBufs );
		//swap display buffers and copy image
		AmUsbFileIoctlSwitchCase( AmUsbIoctlNoSwapBufsCp, AmUsbFileIoctlSwapBufsCp );
		//load sector buffer
		AmUsbFileIoctlSwitchCaseIn( AmUsbIoctlNoLoadSecBuf, AmUsbFileIoctlLoadSecBuf, tAmUsbLoadSecBufIn );
		//read sector buffer
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoReadSecBuf, AmUsbFileIoctlReadSecBuf, tAmUsbReadSecBufOut );
		//get CF sector count
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetCfSecCnt, AmUsbFileIoctlGetCfSecCnt, tAmUsbGetCfSecCntOut );
		//write CF sector
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoWriteCf, AmUsbFileIoctlWriteCf, tAmUsbWriteCfIn, tAmUsbWriteCfOut );
		//read CF sector
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoReadCf, AmUsbFileIoctlReadCf, tAmUsbReadCfIn, tAmUsbReadCfOut );
		//get internal sector count
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetIfSecCnt, AmUsbFileIoctlGetIfSecCnt, tAmUsbGetIfSecCntOut );
		//read internal sector
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoReadIf, AmUsbFileIoctlReadIf, tAmUsbReadIfIn, tAmUsbReadIfOut );
		//get compact flash flag
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetCfFlag, AmUsbFileIoctlGetCfFlag, tAmUsbGetCfFlagOut );
		//load variables
		AmUsbFileIoctlSwitchCaseIn( AmUsbIoctlNoLoadVars, AmUsbFileIoctlLoadVars, tAmUsbLoadVarsIn );
		//read variables
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoReadVars, AmUsbFileIoctlReadVars, tAmUsbReadVarsOut );
		//load stack
		AmUsbFileIoctlSwitchCaseIn( AmUsbIoctlNoLoadStack, AmUsbFileIoctlLoadStack, tAmUsbLoadStackIn );
		//read stack
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoReadStack, AmUsbFileIoctlReadStack, tAmUsbReadStackOut );
		//set stack pointer
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoSetStackPtr, AmUsbFileIoctlSetStackPtr, tAmUsbSetStackPtrIn, tAmUsbSetStackPtrOut );
		//get stack pointer
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetStackPtr, AmUsbFileIoctlGetStackPtr, tAmUsbGetStackPtrOut );
		//get and set key state
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoGetSetKeys, AmUsbFileIoctlGetSetKeys, tAmUsbGetSetKeysIn, tAmUsbGetSetKeysOut );
		//get and set SNES0 key state
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoGetSetSnes0, AmUsbFileIoctlGetSetSnes0, tAmUsbGetSetSnes0In, tAmUsbGetSetSnes0Out  );
		//get and set SNES1 key state
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoGetSetSnes1, AmUsbFileIoctlGetSetSnes1, tAmUsbGetSetSnes1In, tAmUsbGetSetSnes1Out );
		//set time
		AmUsbFileIoctlSwitchCaseIn( AmUsbIoctlNoSetTime, AmUsbFileIoctlSetTime, tAmUsbSetTimeIn );
		//get time
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetTime, AmUsbFileIoctlGetTime, tAmUsbGetTimeOut );
		//get firmware version
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetFwVersion, AmUsbFileIoctlGetFwVersion, tAmUsbGetFwVersionOut );
		//get ID
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetId, AmUsbFileIoctlGetId, tAmUsbGetIdOut );
		//set ID
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoSetId, AmUsbFileIoctlSetId, tAmUsbSetIdIn, tAmUsbSetIdOut );
		//get output style
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetOutStyle, AmUsbFileIoctlGetOutStyle, tAmUsbGetOutStyleOut );
		//set output style
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoSetOutStyle, AmUsbFileIoctlSetOutStyle, tAmUsbSetOutStyleIn, tAmUsbSetOutStyleOut );
		//get output timing
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetOutTiming, AmUsbFileIoctlGetOutTiming, tAmUsbGetOutTimingOut );
		//set output timing
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoSetOutTiming, AmUsbFileIoctlSetOutTiming, tAmUsbSetOutTimingIn, tAmUsbSetOutTimingOut );
		//get version information
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetVersion, AmUsbFileIoctlGetVersion, tAmUsbGetVersionOut );
		//set mode
		AmUsbFileIoctlSwitchCaseInOut( AmUsbIoctlNoSetMode, AmUsbFileIoctlSetMode, tAmUsbSetModeIn, tAmUsbSetModeOut );
		//get mode
		AmUsbFileIoctlSwitchCaseOut( AmUsbIoctlNoGetMode, AmUsbFileIoctlGetMode, tAmUsbGetModeOut );
	} //switch( IoctlNo )

	//unknown ioctl

	//debug message
	AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbFileIoctl: unknown ioctl %08X on file %p\r\n", IoctlNo, pFileObject ) );

	//return error
	*pOutLen = 0;
	return STATUS_ORDINAL_NOT_FOUND; //ERROR_INVALID_ORDINAL
}

//an userspace process is opening one of the devices
NTSTATUS AmUsbFileOpen( struct tAmUsbDev * pDev, PFILE_OBJECT pFileObject )
{
	KIRQL OldIrql;
	struct tAmUsbFilePriv * pFilePriv;
	int FirstUser;
	struct tAmUsbCommBlock * pCommBlock;

	//no private device data
	if( pDev == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );
	//device was unplugged
	if( pDev->Removed )
	{
		//unlock private device data
		//(this restores the original IRQL afterwards)
		KeReleaseSpinLock( &pDev->Lock, OldIrql );
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		//error
		return STATUS_DEVICE_DOES_NOT_EXIST; //ERROR_DEV_NOT_EXIST
	}

	//create private file data
	pFilePriv = (struct tAmUsbFilePriv *)ExAllocatePool( NonPagedPool, sizeof( struct tAmUsbFilePriv ) );
	if( pFilePriv == NULL )
	{
		//debug message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbFileOpen: ExAllocatePool( NonPagedPool, %d ) returned NULL\r\n", sizeof( struct tAmUsbFilePriv ) ) );
		//unlock private device data
		//(this restores the original IRQL afterwards)
		KeReleaseSpinLock( &pDev->Lock, OldIrql );
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		//error
		return STATUS_SECTION_NOT_EXTENDED; //ERROR_OUTOFMEMORY
	}

	//initialize private file data
	//(pDev and pNextFilePriv are initialized a few lines down)
	//use default mode of device as file mode
	pFilePriv->Mode = pDev->Mode;

	//increment our usage count
	FirstUser = pDev->OpenCnt == 0; //remember if this is our first user
	pDev->OpenCnt++;

	//save pointer to private file data in file
	pFileObject->FsContext = (void *)pFilePriv;
	//save pointer to device in the file's private structure
	pFilePriv->pDev = pDev;
	//put private file data into list in device structure
	//the list stays consistent all the time
	//(reading in interrupt is allowed when using only p_next or only p_prev)
	pFilePriv->pPrev = pDev->FilePrivList.pPrev;
	pFilePriv->pNext = &pDev->FilePrivList;
	pFilePriv->pNext->pPrev = pFilePriv;
	pFilePriv->pPrev->pNext = pFilePriv;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbFileOpen: opened file %08X\r\n", pFileObject ) );

	//unlock private device data
	//(this restores the original IRQL afterwards)
	KeReleaseSpinLock( &pDev->Lock, OldIrql );

	//stop command execution if first user is opening
	if( FirstUser )
	{
		pCommBlock = AmUsbCommBlockNew( 1, "\x01", "\x01" );
		if( pCommBlock != NULL )
		{
			pCommBlock->Steps[0].pOutBuffer[0] = 0x00;
			pCommBlock->Steps[0].pFilterBuffer[0] = 0x00;
			pCommBlock->Steps[0].pMaskBuffer[0] = 0xFF;
			AmUsbCommBlockDo( pDev, pCommBlock );
			AmUsbCommBlockFree( pCommBlock );
		}
	}

	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//success
	return STATUS_SUCCESS;
}

//an userspace process is closing one of the devices
//returns 1 on success, 0 on error
NTSTATUS AmUsbFileRelease( PFILE_OBJECT pFileObject )
{
	struct tAmUsbFilePriv * pFilePriv;
	struct tAmUsbDev * pDev;
	KIRQL OldIrql;
	struct tAmUsbCommBlock * pCommBlock;
	int LastUser;

	//get private file data
	pFilePriv = (struct tAmUsbFilePriv *)pFileObject->FsContext;
	//no private file data
	if( pFilePriv == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE
	//get private device data
	pDev = pFilePriv->pDev;
	//no pointer to device data
	if( pDev == NULL )
		//error
		return STATUS_UNSUCCESSFUL; //ERROR_GEN_FAILURE

	//lock device for file operations
	ExAcquireFastMutex( &pDev->FileLock );

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );

	//remove private file data from list in device structure
	//the list stays consistent all the time
	//(reading in interrupt is allowed when using only p_next or only p_prev)
	pFilePriv->pNext->pPrev = pFilePriv->pPrev;
	pFilePriv->pPrev->pNext = pFilePriv->pNext;

	//free private file data
	ExFreePool( pFilePriv );

	//check if device is open
	if( pDev->OpenCnt <= 0 )
	{
		//unlock private device data
		//(this restores the original IRQL afterwards)
		KeReleaseSpinLock( &pDev->Lock, OldIrql );
		//unlock device for file operations
		ExReleaseFastMutex( &pDev->FileLock );
		//error
		return STATUS_DEVICE_DOES_NOT_EXIST; //ERROR_DEV_NOT_EXIST
	}

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "AmUsbFileRelease: closed file %08X\r\n", pFileObject ) );

	//decrement usage count
	pDev->OpenCnt--;
	LastUser = pDev->OpenCnt == 0; //remember if this is our last user

	//device was unplugged
	if( pDev->Removed )
	{
		//device is opened by someone else
		if( pDev->OpenCnt > 0 )
		{
			//unlock private device data
			//(this restores the original IRQL afterwards)
			KeReleaseSpinLock( &pDev->Lock, OldIrql );
			//unlock device for file operations
			ExReleaseFastMutex( &pDev->FileLock );
		}
		//device is not opened by anyone else
		else
		{
			pDev->OpenCnt = 0; //no negative values, please
			//unlock private device data
			//(this restores the original IRQL afterwards)
			KeReleaseSpinLock( &pDev->Lock, OldIrql );
			//unlock device for file operations
			ExReleaseFastMutex( &pDev->FileLock );
			//free device
			AmUsbFree( pDev );
		}
		//success
		return STATUS_SUCCESS;
	}

	//unlock private device data
	//(this restores the original IRQL afterwards)
	KeReleaseSpinLock( &pDev->Lock, OldIrql );

	//start command execution if last user is closing
	if( LastUser )
	{
		pCommBlock = AmUsbCommBlockNew( 2, "\x07\x01", "\x07\x01" );
		if( pCommBlock != NULL )
		{
			memcpy( pCommBlock->Steps[0].pOutBuffer, "\x08\0\0\0\0\x01\xFF", 7 ); //set command pointer to sector 0 offset 0x1FF
			memcpy( pCommBlock->Steps[0].pFilterBuffer, "\x08\0\0\0\0\0\0", 7 );
			memcpy( pCommBlock->Steps[0].pMaskBuffer, "\xFF\0\0\0\0\0\0", 7 );
			pCommBlock->Steps[1].pOutBuffer[0] = 0x01; //start command execution
			pCommBlock->Steps[1].pFilterBuffer[0] = 0x01;
			pCommBlock->Steps[1].pMaskBuffer[0] = 0xFF;
			AmUsbCommBlockDo( pDev, pCommBlock );
			AmUsbCommBlockFree( pCommBlock );
		}
	}
	
	//unlock device for file operations
	ExReleaseFastMutex( &pDev->FileLock );

	//success
	return STATUS_SUCCESS;
}



//*** USB device operations

//device is attached to USB
//sets the pDev member of pDeviceExtension to a new device in case of success
//nothing is done, if pDev is not NULL on entry
//returns -1 on error, 0 on success, 1 if the device is not a AmUsb device, 2 if pDev is already available
int AmUsbProbe( PDEVICE_EXTENSION pDeviceExtension )
{
	PUSB_DEVICE_DESCRIPTOR pDeviceDescriptor;
	PUSBD_INTERFACE_INFORMATION pInterface;
	struct tAmUsbDev * pDev;
	int EndpointsOk;
	KIRQL OldIrql;

	AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbProbe\r\n" ) );

	//raise IRQL to ensure that nobody steals our descriptors while we are processing them
	KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );

	//no device extension available
	if( pDeviceExtension == NULL )
	{
		KeLowerIrql( OldIrql ); //restore old IRQL
		AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbProbe: no device extension!!!\r\n" ) );
		return -1;
	}

	//private device data already available
	if( pDeviceExtension->pDev != NULL )
	{
		KeLowerIrql( OldIrql ); //restore old IRQL
		AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbProbe: private device data already available\r\n" ) );
		return 2;
	}

	//get pointers to device descriptor and to interface
	pDeviceDescriptor = pDeviceExtension->pUsbDeviceDescriptor;
	if( pDeviceDescriptor == NULL )
	{
		KeLowerIrql( OldIrql ); //restore old IRQL
		AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbProbe: no device descriptor!!!\r\n" ) );
		return -1;
	}

	//check if vendor id and product id match
	if( pDeviceDescriptor->idVendor != AmUsbIdVendor || pDeviceDescriptor->idProduct != AmUsbIdProduct )
	{
		KeLowerIrql( OldIrql ); //restore old IRQL
		//device is not a AmUsb device that can be handled
		//(this is not an error, it's just not our device)
		AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbProbe: not an ARCADEmini (vendor or product id)\r\n" ) );
		return 1;
	}

	//restore old IRQL
	KeLowerIrql( OldIrql );

	//now, we are can assume that the device is a AmUsb-device
	//(vendor and product id and high part of revicion match)
	AmUsbDebug( DBGLVL_HIGH, ( "AmUsbProbe: device seems to be an ARCADEmini\r\n" ) );

	//allocate buffer for private data
	pDev = (struct tAmUsbDev *)ExAllocatePool( NonPagedPool, sizeof( struct tAmUsbDev ) );
	//error
	if( pDev == NULL )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbProbe: could not allocate buffer for private data\r\n" ) );
		//abort function with error
		return -1;
	}

	//initialize lock for file operations
	ExInitializeFastMutex( &pDev->FileLock );
	//initialize lock
	KeInitializeSpinLock( &pDev->Lock );
	//initialize pointer to device extension
	pDev->pDeviceExtension = pDeviceExtension;
	//unload process not yet started
	pDev->Unload = 0;
	//device is not yet removed
	pDev->Removed = 0;
	//bulk out/in not active
	pDev->OutActive = 0;
	pDev->InActive = 0;
	//no IRPs, URBs or buffers yet
	pDev->pOutIrp = NULL;
	pDev->pInIrp = NULL;
	pDev->pOutUrb = NULL;
	pDev->pInUrb = NULL;
	pDev->pOutBuffer = NULL;
	pDev->pInBuffer = NULL;
	//initialize DPC object (used to queue some tasks that cannot be done in interrupt)
	KeInitializeDpc( &pDev->DpcObject, AmUsbTask, pDev ); //context is pointer to private device data
	//no task needs to be done yet
	pDev->TaskInStart = 0;
	pDev->TaskCommBlockSend = 0;
	pDev->TaskCommBlockRecv = 0;
	//device is not opened yet
	pDev->OpenCnt = 0;
	//no files have opened this device
	pDev->FilePrivList.pNext = &pDev->FilePrivList;
	pDev->FilePrivList.pPrev = &pDev->FilePrivList;
	//no communication blocks yet
	pDev->CommBlockList.pNext = &pDev->CommBlockList;
	pDev->CommBlockList.pPrev = &pDev->CommBlockList;
	pDev->CommBlockList.StepCnt = 0;
	pDev->CommBlockList.Steps = NULL;
	pDev->CommBlockList.OutStep = 0;
	pDev->CommBlockList.InStep = 0;
	pDev->CommBlockList.pWaitEvent = NULL;
	pDev->CommBlockList.FreeFlag = 0;
	//initialize timeout timer (not active)
	pDev->TimeoutTimerActive = 0;
	KeInitializeTimer( &pDev->TimeoutTimer );
	KeInitializeDpc( &pDev->TimeoutDpc, AmUsbTimeout, pDev ); //context is pointer to private device data
	//default mode is null device
	pDev->Mode = AmUsbModeNull;

	//check endpoints, default is not okay
	EndpointsOk = 0;
	//raise IRQL to ensure that nobody steals our interface while we are processing it
	KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );
	//get interface
	pInterface = pDeviceExtension->pUsbInterface;
	//interface found and pipe array available
	if( pInterface != NULL && pInterface->Pipes != NULL )
	{
		//check endpoints: we want: bulk out, bulk in
		if( pInterface->NumberOfPipes == 2 ) //exactly 2 endpoints
		{
			EndpointsOk = 1;
			//bulk out
			if( (pInterface->Pipes[0].PipeType & 0x03) == 0x02 //bulk
				&& (pInterface->Pipes[0].EndpointAddress & 0x80) == 0x00 ) //out
			{
				//get bulk out parameters
				pDev->OutEndpointAddr = pInterface->Pipes[0].EndpointAddress & 0x7F;
				pDev->OutMaxPacketSize = pInterface->Pipes[0].MaximumPacketSize;
				pDev->OutPipeHandle = pInterface->Pipes[0].PipeHandle;
			}
			else
			{
				//endpoints are not okay
				EndpointsOk = 0;
			}
			//bulk in
			if( (pInterface->Pipes[1].PipeType & 0x03) == 0x02 //bulk
				&& (pInterface->Pipes[1].EndpointAddress & 0x80) != 0x00 ) //in
			{
				//get status in parameters
				pDev->InEndpointAddr = pInterface->Pipes[1].EndpointAddress & 0x7F;
				pDev->InMaxPacketSize = pInterface->Pipes[1].MaximumPacketSize;
				pDev->InPipeHandle = pInterface->Pipes[1].PipeHandle;
			}
			else
			{
				//endpoints are not okay
				EndpointsOk = 0;
			}
		}
	}
	//restore old IRQL
	KeLowerIrql( OldIrql );
	//endpoints are not okay
	if( !EndpointsOk )
	{
		//print error message
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbProbe: device has got unexpected endpoints\r\n" ) );
		//free device
		AmUsbFree( pDev );
		//device is not a AmUsb device
		//(this is not an error, it's just not our device)
		AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbProbe: not an ARCADEmini (endpoints)\r\n" ) );
		return 1;
	}

	//now, we can be sure that device ia a AmUsb-device
	//(endpoints match, too)
	AmUsbDebug( DBGLVL_HIGH, ( "AmUsbProbe: device is an ARCADEmini\r\n" ) );

	//save pointer to private device data in device extension
	pDeviceExtension->pDev = pDev;

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );

	//start bulk input (must be called with device data locked)
	AmUsbInStart( pDev );

	//unlock private device data
	//(this restores the original IRQL afterwards)
	KeReleaseSpinLock( &pDev->Lock, OldIrql );

	//print info message
	AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbProbe: device attached (%08lX)\r\n", pDev ) );

	//return success
	AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbProbe\r\n" ) );
	return 0;
}

//device is removed from USB
//pDev member of pDeviceExtension is set to NULL
//pDeviceExtension member of pDev is set to NULL
//Removed member of pDev is set to 1 if private device data cannot be freed imediately
//nothing happens if pDev member of pDeviceExtension is already NULL
void AmUsbDisconnect( PDEVICE_EXTENSION pDeviceExtension )
{
	struct tAmUsbDev * pDev;
	KIRQL OldIrql;

	AmUsbDebug( DBGLVL_HIGH, ( "entering AmUsbDisconnect\r\n" ) );

	//raise IRQL to ensure that nobody steals our descriptors while we are processing them
	KeRaiseIrql( DISPATCH_LEVEL, &OldIrql );

	//no device extension available
	if( pDeviceExtension == NULL )
	{
		KeLowerIrql( OldIrql ); //restore old IRQL
		AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbDisconnect: no device extension!!!\r\n" ) );
		return;
	}

	//get pointer to private device data
	pDev = pDeviceExtension->pDev;
	if( pDev == NULL )
	{
		KeLowerIrql( OldIrql ); //restore old IRQL
		AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbDisconnect - device extension already gone\r\n" ) );
		return;
	}

	//remove pointer to private device data in pDeviceExtension
	pDeviceExtension->pDev = NULL;
	//remove pointer to device extension in pDev
	pDev->pDeviceExtension = NULL;

	//restore old IRQL
	KeLowerIrql( OldIrql );

	//print info message
	AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbDisconnect: device disconnected (%08lX)\r\n", pDev ) );

	//lock private device data
	//(this first raises IRQL to DISPATCH_LEVEL)
	KeAcquireSpinLock( &pDev->Lock, &OldIrql );

	//cancel bulk input/output
	AmUsbOutCancel( pDev );
	AmUsbInCancel( pDev );

	//unschedule task for things that cannot be done in interrupt
	KeRemoveQueueDpc( &pDev->DpcObject );

	//device is opened by someone
	if( pDev->OpenCnt > 0 )
	{
		//device is now removed
		pDev->Removed = 1;
		//unlock private device data
		//(this restores the original IRQL afterwards)
		KeReleaseSpinLock( &pDev->Lock, OldIrql );
	}
	//device is not opened by anyone
	else
	{
		//unlock private device data
		//(this restores the original IRQL afterwards)
		KeReleaseSpinLock( &pDev->Lock, OldIrql );
		//free device
		AmUsbFree( pDev );
	}

	AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbDisconnect\r\n" ) );
}
