/* ARCADEmini USB driver for Windows
 * version 1.1.0 date 2005-02-10
 * part of the ARCADEmini project - http[s]://arcademini.schuermans.info/
 * Copyright (C) 2003-2005 Stefan Schuermans <1stein@schuermans.info>
 * Copyleft: GNU public license - http://www.gnu.org/copyleft/gpl.html
 */

#include <wdm.h>
#include <stdarg.h>
#include <stdio.h>
#include <usbdi.h>
#include <usbdlib.h>

#include "AmUsbInternal.h"

//Check device extension status flags; 
//Can't accept a new io request if device:
// 1) is removed, 
// 2) has never been started, 
// 3) is stopped,
// 4) has a remove request pending, or
// 5) has a stop device pending
//pDeviceObject: pointer to the device object for this instance of the device
//return value: TRUE if can accept new io requests, else FALSE
BOOLEAN AmUsbCanAcceptIoRequests( IN PDEVICE_OBJECT pDeviceObject )
{
	PDEVICE_EXTENSION pDeviceExtension;
	BOOLEAN fCan = FALSE;

	pDeviceExtension = pDeviceObject->DeviceExtension;

	//flag set when processing IRP_MN_REMOVE_DEVICE
	if ( !pDeviceExtension->DeviceRemoved &&
		 //device must be started( enabled )
		 pDeviceExtension->DeviceStarted &&
 		 //flag set when driver has answered success to IRP_MN_QUERY_REMOVE_DEVICE
		 !pDeviceExtension->RemoveDeviceRequested &&
		 //flag set when driver has answered success to IRP_MN_QUERY_STOP_DEVICE
		 !pDeviceExtension->StopDeviceRequested )
	{
			fCan = TRUE;
	}

	AmUsbDebugCond( DBGLVL_MAXIMUM, !fCan, ( "FALSE return from AmUsbCanAcceptIoRequests()\r\n" ) );

	return fCan;
}

//This is the IRP_MJ_READ routine set in our dispatch table;
//ReadFile() calls from user mode ultimately land here
//pDeviceObject: pointer to the device object for this instance of the device
//pIrp: pointer to the IRP
//return value: NT status code
NTSTATUS AmUsbRead( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	NTSTATUS Status = STATUS_SUCCESS;
	PFILE_OBJECT pFileObject;
	PIO_STACK_LOCATION pIrpStack;
	PDEVICE_EXTENSION pDeviceExtension = pDeviceObject->DeviceExtension;
	unsigned int Cnt;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "entering AmUsbRead()\r\n" ) );

	pIrp->IoStatus.Information = 0;
	//can't accept a new io request if:
	// 1) device is removed, 
	// 2) has never been started, 
	// 3) is stopped,
	// 4) has a remove request pending,
	// 5) has a stop device pending
	if( !AmUsbCanAcceptIoRequests( pDeviceObject ) )
	{
		Status = STATUS_DELETE_PENDING;
		pIrp->IoStatus.Status = Status;
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbRead() can't accept requests, returning STATUS_INSUFFICIENT_RESOURCES\r\n" ) );
		IoCompleteRequest( pIrp, IO_NO_INCREMENT );
		return Status;
	}

	//get pointer to IRP stack, file object and pipe handle
	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	pFileObject = pIrpStack->FileObject;

	//execute the read from user mode
	Status = AmUsbFileRead( pFileObject, pIrp->AssociatedIrp.SystemBuffer, pIrpStack->Parameters.Read.Length, &Cnt );
	pIrp->IoStatus.Status = Status;
	pIrp->IoStatus.Information = Cnt; //number of bytes read

	//complete the request here because we created no staged IRPs
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbRead() with Status 0x%x decimal %d\r\n", Status, Status ) );

	return Status;
}

//This is the IRP_MJ_WRITE routine set in our dispatch table;
//WriteFile() calls from user mode ultimately land here
//pDeviceObject: pointer to the device object for this instance of the device
//pIrp: pointer to the IRP
//return value: NT status code
NTSTATUS AmUsbWrite( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	NTSTATUS Status = STATUS_SUCCESS;
	PFILE_OBJECT pFileObject;
	PIO_STACK_LOCATION pIrpStack;
	PDEVICE_EXTENSION pDeviceExtension = pDeviceObject->DeviceExtension;
	unsigned int Cnt;

	//debug message
	AmUsbDebug( DBGLVL_MEDIUM, ( "entering AmUsbWrite()\r\n" ) );

	pIrp->IoStatus.Information = 0;
	//can't accept a new io request if:
	// 1) device is removed, 
	// 2) has never been started, 
	// 3) is stopped,
	// 4) has a remove request pending,
	// 5) has a stop device pending
	if( !AmUsbCanAcceptIoRequests( pDeviceObject ) )
	{
		Status = STATUS_DELETE_PENDING;
		pIrp->IoStatus.Status = Status;
		AmUsbDebug( DBGLVL_DEFAULT, ( "AmUsbWrite() can't accept requests, returning STATUS_INSUFFICIENT_RESOURCES\r\n" ) );
		IoCompleteRequest( pIrp, IO_NO_INCREMENT );
		return Status;
	}

	//get pointer to IRP stack, file object and pipe handle
	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	pFileObject = pIrpStack->FileObject;

	//execute the write from user mode
	Status = AmUsbFileWrite( pFileObject, pIrp->AssociatedIrp.SystemBuffer, pIrpStack->Parameters.Write.Length, &Cnt );
	pIrp->IoStatus.Status = Status;
	pIrp->IoStatus.Information = Cnt; //number of bytes written

	//complete the request here because we created no staged IRPs
	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	AmUsbDebug( DBGLVL_HIGH, ( "exiting AmUsbWrite() with Status 0x%x decimal %d\r\n", Status, Status ) );

	return Status;
}

//Dispatch table handler for IRP_MJ_DEVICE_CONTROL; 
//Handle DeviceIoControl() calls  from User mode
//pDeviceObject: pointer to the FDO for this instance of the device
//return value: NT status code
NTSTATUS AmUsbIoctl( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	PIO_STACK_LOCATION pIrpStack;
	PVOID pIoBuffer;
	unsigned int InputBufferLength, OutputBufferLength, OutputLength;
	PDEVICE_EXTENSION pDeviceExtension;
	ULONG IoControlCode;
	PFILE_OBJECT pFileObject;
	NTSTATUS Status;

	AmUsbDebug( DBGLVL_MEDIUM, ( "IRP_MJ_DEVICE_CONTROL\r\n" ) );

	AmUsbIncIoCount( pDeviceObject );

	//Get a pointer to the current location in the Irp. This is where
	//the function codes and parameters are located.

	pDeviceExtension = pDeviceObject->DeviceExtension;

	//Can't accept a new io request if:
	// 1) device is removed, 
	// 2) has never been started, 
	// 3) is stopped,
	// 4) has a remove request pending,
	// 5) has a stop device pending
	if( !AmUsbCanAcceptIoRequests( pDeviceObject ) )
	{
		Status = STATUS_DELETE_PENDING;
		pIrp->IoStatus.Status = Status;
		pIrp->IoStatus.Information = 0;

		IoCompleteRequest( pIrp, IO_NO_INCREMENT );

		AmUsbDecIoCount( pDeviceObject );
		return Status;
	}

	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );

	pIrp->IoStatus.Status = STATUS_SUCCESS;
	pIrp->IoStatus.Information = 0;

	//get pointers and lengths of the caller's (user's) IO buffer
	pIoBuffer = pIrp->AssociatedIrp.SystemBuffer;
	InputBufferLength = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;
	OutputBufferLength = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
	IoControlCode = pIrpStack->Parameters.DeviceIoControl.IoControlCode;
	pFileObject = pIrpStack->FileObject;

	//handle IOCTLs from user mode
	Status = AmUsbFileIoctl( pFileObject, IoControlCode, pIoBuffer, InputBufferLength, OutputBufferLength, &OutputLength );
	pIrp->IoStatus.Status = Status;
	pIrp->IoStatus.Information = OutputLength; //set length of returned data

	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	AmUsbDecIoCount( pDeviceObject );

	return Status;
}

//This is the dispatch table routine for IRP_MJ_CLOSE.
//It handles user mode CloseHandle() calls for a pipe
//It closes the File Object for the pipe handle it represents.
//pDeviceObject: pointer to our FDO (Functional Device Object)
//return value:	NT status code
NTSTATUS AmUsbClose( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	NTSTATUS Status;
	NTSTATUS ActStat;
	PFILE_OBJECT pFileObject;
	PIO_STACK_LOCATION pIrpStack;
	PDEVICE_EXTENSION pDeviceExtension;
	PUSBD_PIPE_INFORMATION hPipe = NULL;

	AmUsbDebug( DBGLVL_DEFAULT, ( "entering AmUsbClose\r\n" ) );
	
	AmUsbIncIoCount( pDeviceObject );

	pDeviceExtension = pDeviceObject->DeviceExtension;
	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	pFileObject = pIrpStack->FileObject;

	//free private file data
	Status = AmUsbFileRelease( pFileObject );

	AmUsbDecIoCount( pDeviceObject );
	pIrp->IoStatus.Status = Status;
	pIrp->IoStatus.Information = 0;

	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	//now this device is opened by one less app
	pDeviceExtension->AppOpenCnt--;

	//try to power down device if this is the last app
	ActStat = AmUsbSelfSuspendOrActivate( pDeviceObject, TRUE );

	AmUsbDebug( DBGLVL_DEFAULT, ("exiting AmUsbClose(), AppOpenCnt = decimal %d, status 0x%x\r\n", pDeviceExtension->AppOpenCnt, Status ) );

	return Status;
}

//This is the dispatch table routine for IRP_MJ_CREATE.
//It's the entry point for CreateFile() calls
//user mode apps may open "<name genned fron GUID>.\yy"
//where yy is the internal pipe id
//pDeviceObject: pointer to our FDO (Functional Device Object)
//return value: NT status code
NTSTATUS AmUsbCreate( IN PDEVICE_OBJECT pDeviceObject, IN PIRP pIrp )
{
	NTSTATUS Status = STATUS_SUCCESS;
	PFILE_OBJECT pFileObject;
	PIO_STACK_LOCATION pIrpStack;
	PDEVICE_EXTENSION pDeviceExtension;

	pDeviceExtension = pDeviceObject->DeviceExtension;

	AmUsbDebug( DBGLVL_DEFAULT, ( "entering AmUsbCreate\r\n" ) );

	AmUsbIncIoCount( pDeviceObject );

	//can't accept a new io request if:
	// 1) device is removed, 
	// 2) has never been started, 
	// 3) is stopped,
	// 4) has a remove request pending,
	// 5) has a stop device pending
	if( !AmUsbCanAcceptIoRequests( pDeviceObject ) )
	{
		Status = STATUS_DELETE_PENDING;
		pIrp->IoStatus.Status = Status;
		pIrp->IoStatus.Information = 0;

		IoCompleteRequest( pIrp, IO_NO_INCREMENT );

		AmUsbDecIoCount( pDeviceObject );
		
		AmUsbDebug( DBGLVL_DEFAULT, ( "ABORTING AmUsbCreate\r\n" ) );
		return Status;
	}
	
	pIrpStack = IoGetCurrentIrpStackLocation( pIrp );
	pFileObject = pIrpStack->FileObject;

	//now this device is opened by one more app
	pDeviceExtension->AppOpenCnt++;

	//try to power up device
	AmUsbSelfSuspendOrActivate( pDeviceObject, FALSE );

	//allocate private file data
	Status = AmUsbFileOpen( pDeviceExtension->pDev, pFileObject );
	if( ! NT_SUCCESS( Status ) )
	{
		//now this device is opened by one less app
		pDeviceExtension->AppOpenCnt--;
	}

	pIrp->IoStatus.Status = Status;
	pIrp->IoStatus.Information = 0;

	IoCompleteRequest( pIrp, IO_NO_INCREMENT );

	AmUsbDecIoCount( pDeviceObject );

	AmUsbDebug( DBGLVL_DEFAULT, ( "exiting AmUsbCreate with status 0x%x\r\n", Status ) );

	return Status;
}
